#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - Schema processor."""

import logging
import re
from types import SimpleNamespace
from typing import Any, Optional, Tuple

import voluptuous as vol

from .const import ALL_DEVICE_ID as DEVICE_ID_REGEX
from .const import (
    ATTR_CONTROLLER,
    ATTR_DEVICES,
    ATTR_DHW_SENSOR,
    ATTR_DHW_VALVE,
    ATTR_DHW_VALVE_HTG,
    ATTR_HTG_CONTROL,
)
from .const import ATTR_STORED_HW as ATTR_DHW_SYSTEM
from .const import ATTR_UFH_HTG as ATTR_UFH_SYSTEM
from .const import (
    ATTR_ZONE_IDX,
    ATTR_ZONE_SENSOR,
    ATTR_ZONE_TYPE,
    ATTR_ZONES,
    CTL_DEVICE_ID,
    DEFAULT_MAX_ZONES,
    DHW_SENSOR_ID,
    GWY_DEVICE_ID,
    HTG_DEVICE_ID,
    RLY_DEVICE_ID,
    UFC_DEVICE_ID,
    ZON_SENSOR_ID,
    ZONE_TYPE_SLUGS,
    SystemType,
    __dev_mode__,
)
from .const import id_to_address as addr
from .devices import DEVICE_BY_CLASS_ID

# schema attrs
ATTR_DEVICE_ID = "device_id"
ATTR_HTG_SYSTEM = "system"
ATTR_ORPHANS = "orphans"
ATTR_UFH_CTL = "ufh_controller"

DEVICE_ID = vol.Match(DEVICE_ID_REGEX)
SENSOR_ID = vol.Match(ZON_SENSOR_ID)
CTL_DEVICE_ID = vol.Match(CTL_DEVICE_ID)
DHW_SENSOR_ID = vol.Match(DHW_SENSOR_ID)
GWY_DEVICE_ID = vol.Match(GWY_DEVICE_ID)
HTG_DEVICE_ID = vol.Match(HTG_DEVICE_ID)
RLY_DEVICE_ID = vol.Match(RLY_DEVICE_ID)
UFC_DEVICE_ID = vol.Match(UFC_DEVICE_ID)

ZONE_TYPE_SLUGS = list(ZONE_TYPE_SLUGS)

DOMAIN_ID = vol.Match(r"^[0-9A-F]{2}$")
UFH_IDX_REGEX = r"^0[0-8]$"
UFH_IDX = vol.Match(UFH_IDX_REGEX)
ZONE_IDX = vol.Match(r"^0[0-9AB]$")  # TODO: what if > 12 zones? (e.g. hometronics)

SERIAL_PORT = "serial_port"
PORT_NAME = "port_name"
INPUT_FILE = "input_file"

# Config parameters
DEBUG_MODE = "debug_mode"
ALLOW_LIST = "allow_list"
BLOCK_LIST = "block_list"

CONFIG = "config"
DISABLE_DISCOVERY = "disable_discovery"
DISABLE_SENDING = "disable_sending"
ENABLE_EAVESDROP = "enable_eavesdrop"
ENFORCE_ALLOWLIST = f"enforce_{ALLOW_LIST}"
ENFORCE_BLOCKLIST = f"enforce_{BLOCK_LIST}"
EVOFW_FLAG = "evofw_flag"
MAX_ZONES = "max_zones"
REDUCE_PROCESSING = "reduce_processing"
SERIAL_CONFIG = "serial_config"
USE_NAMES = "use_names"  # use friendly device names from allow_list
USE_SCHEMA = "use_schema"

LOG_FILE_NAME = "file_name"
LOG_ROTATE_BYTES = "rotate_bytes"
LOG_ROTATE_COUNT = "rotate_backups"
PACKET_LOG = "packet_log"  # output

# Schema parameters
SCHEMA = "schema"
MAIN_CONTROLLER = "main_controller"

DONT_CREATE_MESSAGES = 3
DONT_CREATE_ENTITIES = 2
DONT_UPDATE_ENTITIES = 1

# 1/3: Schemas for Configuration
SERIAL_CONFIG_SCHEMA = vol.Schema(
    {
        vol.Optional("baudrate", default=115200): vol.All(
            vol.Coerce(int), vol.Any(57600, 115200)
        ),  # NB: HGI80 does not work, except at 115200 - so must be default
        vol.Optional("timeout", default=0): vol.Any(None, int),  # TODO: default None?
        vol.Optional("dsrdtr", default=False): bool,
        vol.Optional("rtscts", default=False): bool,
        vol.Optional("xonxoff", default=True): bool,  # set True to remove \x11
    }
)
SERIAL_PORT_SCHEMA = vol.Schema(
    {
        vol.Required(SERIAL_PORT): vol.Any(
            None,
            str,
            SERIAL_CONFIG_SCHEMA.extend({vol.Required(PORT_NAME): vol.Any(None, str)}),
        )
    }
)

PACKET_LOG_SCHEMA = vol.Schema(
    {
        vol.Required(LOG_FILE_NAME, default=None): vol.Any(None, str),
        vol.Optional(LOG_ROTATE_BYTES, default=None): vol.Any(None, int),
        vol.Optional(LOG_ROTATE_COUNT, default=None): vol.Any(None, int),
    },
    extra=vol.PREVENT_EXTRA,
)

CONFIG_SCHEMA = vol.Schema(
    {
        vol.Optional(DISABLE_DISCOVERY, default=False): bool,
        vol.Optional(DISABLE_SENDING, default=False): bool,
        vol.Optional(ENABLE_EAVESDROP, default=False): bool,
        vol.Optional(REDUCE_PROCESSING, default=0): vol.All(
            int, vol.Range(min=0, max=DONT_CREATE_MESSAGES)
        ),
        vol.Optional(MAX_ZONES, default=DEFAULT_MAX_ZONES): vol.All(
            int, vol.Range(min=1, max=16)
        ),
        vol.Optional(USE_SCHEMA, default=True): vol.Any(None, bool),
        vol.Optional(ENFORCE_ALLOWLIST, default=None): vol.Any(None, bool),
        vol.Optional(ENFORCE_BLOCKLIST, default=None): vol.Any(None, bool),
        vol.Optional(USE_NAMES, default=None): vol.Any(None, bool),
        vol.Optional(EVOFW_FLAG, default=None): vol.Any(None, str),
    },
    extra=vol.ALLOW_EXTRA,  # TODO: remove for production
)

FILTER_SCHEMA = vol.Any([], vol.All(vol.Schema([DEVICE_ID]), vol.Length(min=0)))
KNOWNS_SCHEMA = vol.Schema(
    {
        vol.Optional(DEVICE_ID): vol.Any(
            None,
            {
                vol.Optional("name", default=None): vol.Any(None, str),
                vol.Optional("dev_class"): vol.Any(*DEVICE_BY_CLASS_ID.keys()),
            },  # TODO: override type-by-addr.id with dev_class
        )
    },
    extra=vol.PREVENT_EXTRA,
)

# 2/3: Schemas for Heating systems
ATTR_SYS_PROFILE = "_profile"
SYSTEM_PROFILES = (SystemType.EVOHOME, SystemType.HOMETRONICS, SystemType.SUNDIAL)

HTG_SCHEMA = vol.Schema(
    {
        vol.Required(ATTR_HTG_CONTROL, default=None): vol.Any(None, HTG_DEVICE_ID),
        vol.Optional(ATTR_SYS_PROFILE, default=SystemType.EVOHOME): vol.Any(
            *SYSTEM_PROFILES
        ),
    },
    # extra=vol.ALLOW_EXTRA,  # TODO: remove me
)
DHW_SCHEMA = vol.Schema(
    {
        vol.Optional(ATTR_DHW_SENSOR, default=None): vol.Any(None, DHW_SENSOR_ID),
        vol.Optional(ATTR_DHW_VALVE, default=None): vol.Any(None, RLY_DEVICE_ID),
        vol.Optional(ATTR_DHW_VALVE_HTG, default=None): vol.Any(None, RLY_DEVICE_ID),
    }
)
UFC_CIRCUIT = vol.Schema(
    {
        vol.Required(UFH_IDX): vol.Any(
            {vol.Optional(ATTR_ZONE_IDX): vol.Any(ZONE_IDX)},
        ),
    }
)
UFH_SCHEMA = vol.Schema(
    {
        vol.Required(DEVICE_ID): vol.Any(
            None, {vol.Optional("ufh_circuits"): vol.Any(None, dict)}
        )
    }
)
UFH_SCHEMA = vol.All(UFH_SCHEMA, vol.Length(min=1, max=3))

SENSOR_SCHEMA = vol.Schema(
    {
        vol.Required(ATTR_DEVICE_ID): SENSOR_ID,
        vol.Optional("is_faked"): bool,
        # vol.Optional("native_id"): DEVICE_ID,
    }
)
# SENSOR_SCHEMA = vol.Any(None, SENSOR_ID, SENSOR_SCHEMA)
ZONE_SCHEMA = vol.Schema(
    {
        vol.Optional(ATTR_ZONE_TYPE, default=None): vol.Any(None, ZONE_TYPE_SLUGS),
        vol.Optional(ATTR_ZONE_SENSOR, default=None): vol.Any(
            None, SENSOR_ID, SENSOR_SCHEMA
        ),
        vol.Optional(ATTR_DEVICES, default=[]): vol.Any([], [DEVICE_ID]),
    }
)
ZONE_SCHEMA = vol.Schema({vol.Required(ZONE_IDX): ZONE_SCHEMA})
ZONE_SCHEMA = vol.All(ZONE_SCHEMA, vol.Length(min=1, max=DEFAULT_MAX_ZONES))
SYSTEM_SCHEMA = vol.Schema(
    {
        vol.Required(ATTR_CONTROLLER): CTL_DEVICE_ID,
        vol.Optional(ATTR_HTG_SYSTEM, default={}): vol.Any({}, HTG_SCHEMA),
        vol.Optional(ATTR_DHW_SYSTEM, default={}): vol.Any({}, DHW_SCHEMA),
        vol.Optional(ATTR_UFH_SYSTEM, default={}): vol.Any({}, UFH_SCHEMA),
        vol.Optional(ATTR_ORPHANS, default=[]): vol.Any([], [DEVICE_ID]),
        vol.Optional(ATTR_ZONES, default={}): vol.Any({}, ZONE_SCHEMA),
    },
    extra=vol.ALLOW_EXTRA,  # TODO: remove me - But: Causes an issue?
)
SYSTEM_SCHEMA = vol.Schema(vol.Any({}, SYSTEM_SCHEMA))


# 3/3: Global Schemas
GLOBAL_CONFIG_SCHEMA = vol.Schema(
    {
        vol.Required(CONFIG): CONFIG_SCHEMA.extend(
            {
                vol.Optional(SERIAL_CONFIG): SERIAL_CONFIG_SCHEMA,
                vol.Optional(PACKET_LOG, default={}): vol.Any({}, PACKET_LOG_SCHEMA),
            }
        ),
        vol.Optional(ALLOW_LIST, default={}): vol.All(KNOWNS_SCHEMA, vol.Length(min=0)),
        vol.Optional(BLOCK_LIST, default={}): vol.All(KNOWNS_SCHEMA, vol.Length(min=0)),
    },
    extra=vol.REMOVE_EXTRA,
)


DEV_MODE = __dev_mode__  # or True

_LOGGER = logging.getLogger(__name__)
if DEV_MODE:
    _LOGGER.setLevel(logging.DEBUG)


def load_config_schema(
    serial_port, input_file, **kwargs
) -> Tuple[SimpleNamespace, dict, list, list]:
    """Process the configuration, including any filter lists."""

    schema = {
        k: v for k, v in kwargs.items() if k not in (ALLOW_LIST, BLOCK_LIST, CONFIG)
    }
    kwargs = GLOBAL_CONFIG_SCHEMA(kwargs)

    allow_list = kwargs.pop(ALLOW_LIST)
    block_list = kwargs.pop(BLOCK_LIST)

    config = CONFIG_SCHEMA.extend(
        {vol.Optional(SERIAL_CONFIG, default={}): SERIAL_CONFIG_SCHEMA}
    )(kwargs[CONFIG])

    known_devices = {**allow_list, **block_list}
    if config[USE_NAMES] is None:
        config[USE_NAMES] = known_devices and any(
            v.get("name")
            for v in known_devices.values()
            if v is not None  # if isinstance(v, dict)
        )

    if serial_port and input_file:
        _LOGGER.warning(
            "Serial port was specified (%s), so input file (%s) will be ignored",
            serial_port,
            input_file,
        )
    elif serial_port is None:
        config[DISABLE_SENDING] = True

    if config[DISABLE_SENDING]:
        config[DISABLE_DISCOVERY] = True

    if config[ENABLE_EAVESDROP]:
        _LOGGER.warning(
            f"{ENABLE_EAVESDROP} was enabled: not doing so is strongly recommended"
            " (there be dragons here)"
        )

    if config[ENFORCE_ALLOWLIST] is None:
        config[ENFORCE_ALLOWLIST] = (
            False if config[ENFORCE_BLOCKLIST] else bool(allow_list)
        )
    if config[ENFORCE_BLOCKLIST] is None:
        config[ENFORCE_BLOCKLIST] = (
            False if config[ENFORCE_ALLOWLIST] else bool(block_list)
        )

    assert (
        config[ENFORCE_ALLOWLIST] is not None and config[ENFORCE_BLOCKLIST] is not None
    )

    if config[ENFORCE_ALLOWLIST] and not allow_list:
        _LOGGER.warning(f"An empty {ALLOW_LIST} was enabled, so will be ignored")
        config[ENFORCE_ALLOWLIST] = False
    if config[ENFORCE_BLOCKLIST] and not block_list:
        _LOGGER.warning(f"An empty {BLOCK_LIST} was enabled, so will be ignored")
        config[ENFORCE_BLOCKLIST] = False

    if config[ENFORCE_ALLOWLIST] and config[ENFORCE_BLOCKLIST]:
        _LOGGER.warning(
            f"Both an {ALLOW_LIST} and a {BLOCK_LIST} were enabled, "
            f"so the {BLOCK_LIST} will be ignored"
        )
        config[ENFORCE_BLOCKLIST] = False
    elif not config[ENFORCE_ALLOWLIST]:
        _LOGGER.warning(
            f"No {ALLOW_LIST} was configured, but one is strongly recommended"
        )

    if config[ENFORCE_ALLOWLIST]:
        _LOGGER.debug(f"An {ALLOW_LIST} has been created, length = {len(allow_list)}")
    elif config[ENFORCE_BLOCKLIST]:
        _LOGGER.debug(f"A {BLOCK_LIST} has been created, length = {len(block_list)}")

    # TODO: use dict or list? - will need to add 18? when we see it?
    # allow_list = (
    #     list(allow_list.keys()) if config[ENFORCE_ALLOWLIST] else []
    # )
    # block_list = (
    #     list(block_list.keys()) if config[ENFORCE_BLOCKLIST] else []
    # )

    config = SimpleNamespace(**config)

    return (config, schema, allow_list, block_list)


def _get_device(gwy, dev_addr, ctl_addr=None, **kwargs) -> Optional[Any]:
    """A wrapper to enforce device filters."""
    err_msg = None

    if gwy.config.enforce_allow_list:
        if ctl_addr and ctl_addr.id not in gwy._include:
            err_msg = f"{ctl_addr.id} is in the {SCHEMA}, but not in the {ALLOW_LIST}"
        elif dev_addr.id not in gwy._include:
            err_msg = f"{dev_addr.id} is in the {SCHEMA}, but not in the {ALLOW_LIST}"

    elif gwy.config.enforce_block_list:
        if ctl_addr and ctl_addr.id in gwy._exclude:
            err_msg = f"{ctl_addr.id} is in the {SCHEMA}, but also in the {BLOCK_LIST}"
        elif dev_addr.id not in gwy._include:
            err_msg = f"{dev_addr.id} is in the {SCHEMA}, but also in the {BLOCK_LIST}"

    if err_msg:
        _LOGGER.error(f"%s: check the lists and the (cached) {SCHEMA}", err_msg)

    return gwy._get_device(dev_addr, ctl_addr=ctl_addr, **kwargs)


def load_system_schema(gwy, **kwargs) -> dict:
    """Process the schema, and the configuration and return True if it is valid."""
    # TODO: check a sensor is not a device in another zone

    gwy._clear_state()  # TODO: consider need for this (here, or at all)

    known_devices = kwargs.pop(ALLOW_LIST, {})
    known_devices.update(kwargs.pop(BLOCK_LIST, {}))

    [_get_device(gwy, addr(device_id)) for device_id in kwargs.pop(ATTR_ORPHANS, [])]

    if SCHEMA in kwargs:
        # ctl_id = kwargs.pop(ATTR_CONTROLLER)
        _load_system_schema(gwy, schema=kwargs[SCHEMA])
        gwy.evo = gwy.system_by_id.get(kwargs[SCHEMA][ATTR_CONTROLLER])

    elif kwargs.get(MAIN_CONTROLLER):
        [
            _load_system_schema(gwy, schema)
            for k, schema in kwargs.items()
            if re.match(DEVICE_ID_REGEX, k)
        ]
        gwy.evo = gwy.system_by_id.get(kwargs[MAIN_CONTROLLER])

    return {
        **known_devices,
        **{d: None for d in gwy.device_by_id if d not in known_devices},
    }


def shrink_dict(_dict):
    return {
        k: shrink_dict(dict(v)) if isinstance(v, dict) else v
        for k, v in _dict.items()
        if bool(v)
    }


def _load_system_schema(gwy, schema) -> Tuple[dict, dict]:
    # org_schema = schema
    schema = SYSTEM_SCHEMA(schema)

    ctl_id = schema[ATTR_CONTROLLER]
    profile = schema[ATTR_HTG_SYSTEM].get(ATTR_SYS_PROFILE)
    ctl = _get_device(gwy, addr(ctl_id), ctl_addr=addr(ctl_id), profile=profile)

    # if not ctl:
    #     return

    htg_ctl_id = schema[ATTR_HTG_SYSTEM].get(ATTR_HTG_CONTROL)
    if htg_ctl_id:
        ctl._evo._set_htg_control(_get_device(gwy, addr(htg_ctl_id), ctl_addr=ctl))

    dhw = schema.get(ATTR_DHW_SYSTEM, {})
    if dhw:
        ctl._evo._set_dhw(ctl._evo._get_dhw())

        dhw_sensor_id = dhw.get(ATTR_DHW_SENSOR)
        if dhw_sensor_id:
            ctl._evo.dhw._set_sensor(
                _get_device(gwy, addr(dhw_sensor_id), ctl_addr=ctl)
            )

        dhw_valve_id = dhw.get(ATTR_DHW_VALVE)
        if dhw_valve_id:
            ctl._evo.dhw._set_dhw_valve(
                _get_device(gwy, addr(dhw_valve_id), ctl_addr=ctl)
            )

        htg_valve_id = dhw.get(ATTR_DHW_VALVE_HTG)
        if htg_valve_id:
            ctl._evo.dhw._set_htg_valve(
                _get_device(gwy, addr(htg_valve_id), ctl_addr=ctl)
            )

    for zone_idx, attr in schema[ATTR_ZONES].items():
        zone = ctl._evo._get_zone(zone_idx, zone_type=attr.get(ATTR_ZONE_TYPE))

        sensor_id = attr.get(ATTR_ZONE_SENSOR)
        is_faked = None
        if isinstance(sensor_id, dict):
            is_faked = sensor_id.get("is_faked")
            sensor_id = sensor_id[ATTR_DEVICE_ID]

        if sensor_id:
            zone._set_sensor(
                _get_device(gwy, addr(sensor_id), ctl_addr=ctl, domain_id=zone_idx)
            )  # TODO: use domain_id=zone_idx or not
        if is_faked:
            zone.sensor._make_fake()

        for device_id in attr.get(ATTR_DEVICES, []):
            _get_device(gwy, addr(device_id), ctl_addr=ctl, domain_id=zone_idx)

    # TODO: not create orphans by default?
    orphan_ids = schema.get(ATTR_ORPHANS, [])
    if orphan_ids:
        [_get_device(gwy, addr(device_id), ctl_addr=ctl) for device_id in orphan_ids]

    ufh_ctl_ids = schema.get(ATTR_UFH_SYSTEM, {})
    if ufh_ctl_ids:
        for ufc_id, _ in ufh_ctl_ids.items():
            _ = _get_device(gwy, addr(ufc_id), ctl_addr=ctl)

    if False and DEV_MODE:
        import json

        src = json.dumps(shrink_dict(schema), sort_keys=True)
        dst = json.dumps(shrink_dict(gwy.system_by_id[ctl_id].schema), sort_keys=True)
        # assert dst == src, "They don't match!"
        print(src)
        print(dst)

    return ctl



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - The evohome-compatible zones."""

import logging
from asyncio import Task
from datetime import datetime as dt
from datetime import timedelta as td
from inspect import getmembers, isclass
from sys import modules
from types import SimpleNamespace
from typing import Optional

from .command import Command, Schedule
from .const import (
    _000C_DEVICE,
    ATTR_DEVICES,
    ATTR_DHW_SENSOR,
    ATTR_DHW_VALVE,
    ATTR_DHW_VALVE_HTG,
    ATTR_HEAT_DEMAND,
    ATTR_RELAY_DEMAND,
    ATTR_SETPOINT,
    ATTR_TEMP,
    ATTR_WINDOW_OPEN,
    ATTR_ZONE_SENSOR,
    ATTR_ZONE_TYPE,
    DEVICE_HAS_ZONE_SENSOR,
    DISCOVER_ALL,
    DISCOVER_PARAMS,
    DISCOVER_SCHEMA,
    DISCOVER_STATUS,
    ZONE_CLASS_MAP,
    ZONE_TYPE_MAP,
    ZONE_TYPE_SLUGS,
    ZoneMode,
    __dev_mode__,
)
from .devices import BdrSwitch, Device, DhwSensor, Entity
from .exceptions import CorruptStateError
from .helpers import schedule_task

# from .ramses import RAMSES_ZONES, RAMSES_ZONES_ALL

I_, RQ, RP, W_ = " I", "RQ", "RP", " W"

DEV_MODE = __dev_mode__ and False

_LOGGER = logging.getLogger(__name__)
if DEV_MODE:
    _LOGGER.setLevel(logging.DEBUG)


ZONE_CLASS = SimpleNamespace(
    DHW="DHW",  # Stored HW (not a zone)
    ELE="ELE",  # Electric
    MIX="MIX",  # Mix valve
    RAD="RAD",  # Radiator
    UFH="UFH",  # Underfloor heating
    VAL="VAL",  # Zone valve
)


class ZoneBase(Entity):
    """The Zone/DHW base class."""

    # __zon_class__ = None  # NOTE: this would cause problems

    def __init__(self, evo, zone_idx) -> None:
        _LOGGER.debug("Creating a Zone: %s_%s (%s)", evo, zone_idx, self.__class__)
        super().__init__(evo._gwy)

        self.id, self.idx = f"{evo.id}_{zone_idx}", zone_idx
        self._evo, self._ctl = self._set_system(evo, zone_idx)

        self._name = None
        self._zone_type = None

    def __repr__(self) -> str:
        return f"{self.id} ({self.heating_type})"

    def __lt__(self, other) -> bool:
        if not hasattr(other, "idx"):
            return NotImplemented
        return self.idx < other.idx

    def _set_system(self, parent, zone_idx):
        """Set the zone's parent system, after validating it."""

        # these imports are here to prevent circular references
        from .systems import System

        try:
            if zone_idx != "HW" and int(zone_idx, 16) >= parent.max_zones:
                raise ValueError(f"{self}: invalid zone_idx {zone_idx} (> max_zones")
        except (TypeError, ValueError):
            raise TypeError(f"{self}: invalid zone_idx {zone_idx}")

        if not isinstance(parent, System):
            raise TypeError(f"{self}: parent must be a System, not {parent}")

        if zone_idx != "HW":
            if self.idx in parent.zone_by_idx:
                raise LookupError(f"{self}: duplicate zone_idx: {zone_idx}")
            parent.zone_by_idx[zone_idx] = self
            parent.zones.append(self)

        self._ctl = parent._ctl

        return parent, parent._ctl

    def _handle_msg(self, msg) -> None:
        """Validate packets by verb/code."""
        # TODO: assert msg.src is self, "Devices should only keep msgs they sent"
        super()._handle_msg(msg)

        # if not self._zone_type:
        #     return
        # ramses_zones = RAMSES_ZONES.get(self._zone_type, RAMSES_ZONES_ALL)
        # assert (
        #     msg.code in ramses_zones
        # ), f"Unknown code for {str(self)}: {msg.verb}/{msg.code}"
        # assert (
        #     msg.verb in ramses_zones[msg.code]
        # ), f"Unknown verb for {str(self)}: {msg.verb}/{msg.code}"

    def _send_cmd(self, code, **kwargs) -> None:
        dest = kwargs.pop("dest_addr", self._ctl.id)
        payload = kwargs.pop("payload", f"{self.idx}00")
        super()._send_cmd(code, dest, payload, **kwargs)

    @property
    def heating_type(self) -> Optional[str]:
        """Return the type of the zone/DHW (e.g. electric_zone, stored_dhw)."""
        return self.__zon_class__


class ZoneSchedule:  # 0404  # TODO: add for DHW
    """Evohome zones have a schedule."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._schedule = Schedule(self)

    # def _discover(self, discover_flag=DISCOVER_ALL) -> None:

    #     if discover_flag & DISCOVER_STATUS:  # TODO: add back in
    #         self._loop.create_task(self.get_schedule())  # 0404

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        if msg.code == "0404" and msg.verb == RP:
            _LOGGER.debug("Zone(%s): Received RP/0404 (schedule) pkt", self)

    async def get_schedule(self, force_refresh=None) -> Optional[dict]:
        schedule = await self._schedule.get_schedule(force_refresh=force_refresh)
        if schedule:
            return schedule["schedule"]

    async def set_schedule(self, schedule) -> None:
        schedule = {"zone_idx": self.idx, "schedule": schedule}
        await self._schedule.set_schedule(schedule)

    @property
    def status(self) -> dict:
        return {**super().status, "schedule": self._schedule.schedule.get("schedule")}


class RelayDemand:  # 0008
    """Not all zones call for heat."""

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        super()._discover(discover_flag=discover_flag)

        if discover_flag & DISCOVER_STATUS:
            self._send_cmd("0008")  # , payload=self.idx

    @property
    def relay_demand(self) -> Optional[float]:  # 0008
        if "0008" in self._msgs:
            return self._msgs["0008"].payload[ATTR_RELAY_DEMAND]

    @property
    def status(self) -> dict:
        return {**super().status, ATTR_RELAY_DEMAND: self.relay_demand}


class DhwZone(ZoneBase):  # CS92A  # TODO: add Schedule
    """The DHW class."""

    __zon_class__ = ZONE_CLASS.DHW

    def __init__(
        self, ctl, zone_idx="HW", sensor=None, dhw_valve=None, htg_valve=None
    ) -> None:
        super().__init__(ctl, zone_idx)

        ctl._set_dhw(self)
        # if profile == ZONE_CLASS.DHW and evo.dhw is None:
        #     evo.dhw = zone

        self._dhw_sensor = None
        self._dhw_valve = None
        self._htg_valve = None

        self._zone_type = ZONE_CLASS.DHW

        self._dhw_mode = None
        self._dhw_params = None
        self._dhw_temp = None
        self._relay_demand = None
        self._relay_failsafe = None

        self._heat_demand = None

        if sensor:
            self._set_sensor(sensor)
        if dhw_valve:
            self._set_dhw_valve(dhw_valve)
        if htg_valve:
            self._set_htg_valve(htg_valve)

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        # super()._discover(discover_flag=discover_flag)

        # if False and __dev_mode__ and self.idx == "FA":  # dev/test code
        #     self.async_set_override(state="On")

        if discover_flag & DISCOVER_SCHEMA:
            [  # 000C: find the DHW relay(s), if any, see: _000C_DEVICE_TYPE
                self._send_cmd("000C", payload=dev_type)
                for dev_type in (
                    f"00{_000C_DEVICE.DHW_SENSOR}",
                    f"00{_000C_DEVICE.DHW}",
                    f"01{_000C_DEVICE.DHW}",
                )
            ]

        if discover_flag & DISCOVER_PARAMS:
            self._gwy.send_cmd(
                Command.get_dhw_params(self._ctl.id), period=td(hours=12)
            )

        if discover_flag & DISCOVER_STATUS:
            self._gwy.send_cmd(Command.get_dhw_mode(self._ctl.id), period=td(hours=12))
            self._gwy.send_cmd(
                Command.get_dhw_temp(self._ctl.id), period=td(minutes=30)
            )

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        if msg.code == "0008":
            self._relay_demand = msg

        elif msg.code == "0009":
            self._relay_failsafe = msg

        elif msg.code == "10A0":
            self._dhw_params = msg

        elif msg.code == "1260":
            self._dhw_temp = msg

        elif msg.code == "1F41":
            self._dhw_mode = msg

    def _set_dhw_device(self, new_dev, old_dev, attr_name, dev_class, domain_id):
        if old_dev is new_dev:
            return old_dev
        if old_dev is not None:
            raise CorruptStateError(
                f"{self} changed {attr_name}: {old_dev} to {new_dev}"
            )

        if not isinstance(new_dev, dev_class):
            raise TypeError(f"{self}: {attr_name} can't be {dev_class}")

        new_dev._set_parent(self, domain=domain_id)
        return new_dev

    def _set_sensor(self, device: DhwSensor) -> None:
        """Set the temp sensor for this DHW system (07: only)."""

        # 07:38:39.124 047 RQ --- 07:030741 01:102458 --:------ 10A0 006 00181F0003E4
        # 07:38:39.140 062 RP --- 01:102458 07:030741 --:------ 10A0 006 0018380003E8

        self._dhw_sensor = self._set_dhw_device(
            device, self._dhw_sensor, ATTR_ZONE_SENSOR, DhwSensor, "FA"
        )

    @property
    def sensor(self) -> DhwSensor:  # self._dhw_sensor
        return self._dhw_sensor

    def _set_dhw_valve(self, device: BdrSwitch) -> None:  # self._dhw_valve
        """Set the hotwater valve relay for this DHW system (13: only).

        Check and Verb the DHW sensor (07:) of this system/CTL (if there is one).

        There is only 1 way to eavesdrop a controller's DHW sensor:
        1.  The 10A0 RQ/RP *from/to a 07:* (1x/4h)

        The RQ is initiated by the DHW, so is not authorative (the CTL will RP any RQ).
        The I/1260 is not to/from a controller, so is not useful.
        """  # noqa: D402

        self._dhw_valve = self._set_dhw_device(
            device, self._dhw_valve, ATTR_DHW_VALVE, BdrSwitch, "FA"
        )

    @property
    def hotwater_valve(self) -> BdrSwitch:  # self._dhw_valve
        return self._dhw_valve

    def _set_htg_valve(self, device: BdrSwitch) -> None:  # self._htg_valve
        """Set the heating valve relay for this DHW system (13: only)."""

        self._htg_valve = self._set_dhw_device(
            device, self._htg_valve, ATTR_DHW_VALVE_HTG, BdrSwitch, "F9"
        )

    @property
    def heating_valve(self) -> BdrSwitch:  # self._htg_valve
        return self._htg_valve

    @property
    def name(self) -> str:
        return "Stored HW"

    @property
    def config(self) -> Optional[dict]:  # 10A0
        result = self._msg_payload(self._dhw_params)
        if result:
            return {k: v for k, v in result.items() if k != "dhw_idx"}

    @property
    def mode(self) -> Optional[dict]:  # 1F41
        # active (setpoint), mode, until
        return self._msg_payload(self._dhw_mode)

    @property
    def setpoint(self) -> Optional[float]:  # 1F41
        return self._msg_payload(self._dhw_params, ATTR_SETPOINT)

    @setpoint.setter
    def setpoint(self, value) -> None:  # 1F41
        return self.set_config(setpoint=value)

    @property
    def temperature(self) -> Optional[float]:  # 1260
        return self._msg_payload(self._dhw_temp, ATTR_TEMP)

    @property
    def heat_demand(self) -> Optional[float]:  # 3150
        return self._msg_payload(self._heat_demand, ATTR_HEAT_DEMAND)

    @property
    def relay_demand(self) -> Optional[float]:  # 0008
        return self._msg_payload(self._relay_demand, "relay_demand")

    @property  # only seen with FC, but seems should pair with 0008?
    def relay_failsafe(self) -> Optional[float]:  # 0009
        return self._msg_payload(self._relay_failsafe, "relay_failsafe")

    def set_mode(self, mode=None, active=None, until=None) -> Task:
        """Set the DHW mode (mode, active, until)."""
        cmd = Command.set_dhw_mode(self._ctl.id, mode=mode, active=active, until=until)
        return self._gwy.send_cmd(cmd)

    def set_boost_mode(self) -> Task:
        """Enable DHW for an hour, despite any schedule."""
        return self.set_mode(
            mode=ZoneMode.TEMPORARY, active=True, until=dt.now() + td(hours=1)
        )

    def reset_mode(self) -> Task:  # 1F41
        """Revert the DHW to following its schedule."""
        return self.set_mode(mode=ZoneMode.SCHEDULE)

    def set_config(self, setpoint=None, overrun=None, differential=None) -> Task:
        """Set the DHW parameters (setpoint, overrun, differential)."""
        # if self._dhw_params:  # 10A0
        # if setpoint is None:
        #     setpoint = self._msg_payload(self._dhw_params, ATTR_SETPOINT)
        # if overrun is None:
        #     overrun = self._msg_payload(self._dhw_params, "overrun")
        # if differential is None:
        #     setpoint = self._msg_payload(self._dhw_params, "differential")

        cmd = Command.set_dhw_params(self._ctl.id, setpoint, overrun, differential)
        return self._gwy.send_cmd(cmd)

    def reset_config(self) -> Task:  # 10A0
        """Reset the DHW parameters to their default values."""
        return self.set_config(setpoint=50, overrun=5, differential=1)

    @property
    def schema(self) -> dict:
        """Return the DHW's schema (devices)."""
        return {
            ATTR_DHW_SENSOR: self.sensor.id if self.sensor else None,
            ATTR_DHW_VALVE: self.hotwater_valve.id if self.hotwater_valve else None,
            ATTR_DHW_VALVE_HTG: self.heating_valve.id if self.heating_valve else None,
        }

    @property
    def params(self) -> dict:
        """Return the DHW's configuration (excl. schedule)."""
        return {a: getattr(self, a) for a in ("config", "mode")}

    @property
    def status(self) -> dict:
        """Return the DHW's current state."""
        return {a: getattr(self, a) for a in (ATTR_TEMP, ATTR_HEAT_DEMAND)}


class Zone(ZoneSchedule, ZoneBase):
    """The Zone base class."""

    __zon_class__ = None  # Unknown

    def __init__(self, evo, zone_idx, sensor=None, actuators=None) -> None:
        """Create a zone.

        The type of zone may not be known at instantiation. Even when it is known, zones
        are still created without a type before they are subsequently promoted, so that
        both schemes (e.g. eavesdropping, vs probing) are the same.

        In addition, an electric zone may subsequently turn out to be a zone valve zone.
        """
        if int(zone_idx, 16) >= evo.max_zones:
            raise ValueError(f"Invalid zone idx: {zone_idx} (exceeds max_zones)")

        super().__init__(evo, zone_idx)

        self.devices = []
        self.device_by_id = {}
        self._sensor = None

        self._heat_demand = None  # Not used by ELE
        self._mode = None
        # self._name = None  # from super()
        self._setpoint = None
        self._temperature = None
        self._zone_config = None

        # these needed here, as we can't use those __init__()s
        self._mix_config = None  # MIX
        self._ufh_setpoint = None  # UFH
        self._window_open = None  # RAD (or: ALL)
        self._actuator_state = None  # ELE, ZON

        # self._schedule = Schedule(self)  # TODO:

        if sensor:
            self._set_sensor(sensor)

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        super()._discover(discover_flag=discover_flag)

        # TODO: add code to determine zone type if it doesn't have one, using 0005s
        if discover_flag & DISCOVER_SCHEMA:
            self._send_cmd("000C", payload=f"{self.idx}{_000C_DEVICE.ALL}")
            self._send_cmd("000C", payload=f"{self.idx}{_000C_DEVICE.ALL_SENSOR}")

        if discover_flag & DISCOVER_PARAMS:
            self._gwy.send_cmd(Command.get_zone_config(self._ctl.id, self.idx))
            self._gwy.send_cmd(
                Command.get_zone_name(self._ctl.id, self.idx), period=td(hours=4)
            )

        if discover_flag & DISCOVER_STATUS:  # every 1h, CTL will not respond to a 3150
            self._gwy.send_cmd(
                Command.get_zone_mode(self._ctl.id, self.idx), period=td(minutes=30)
            )
            self._gwy.send_cmd(Command.get_zone_temperature(self._ctl.id, self.idx))
            self._gwy.send_cmd(Command.get_zone_window_state(self._ctl.id, self.idx))

        # start collecting the schedule
        # self._schedule.req_schedule()  # , restart=True) start collecting schedule

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        if isinstance(msg.payload, list):
            assert self.idx in [d["zone_idx"] for d in msg.payload]

        if msg.code == "0004":
            self._name = msg

        # not UFH (it seems), but ELE or VAL; and possibly a MIX support 0008 too
        elif msg.code in ("0008", "0009"):  # TODO: how to determine is/isn't MIX?
            assert msg.src.type in ("01", "13"), msg.src.type  # 01 as a stat
            assert self._zone_type in (None, "ELE", "VAL", "MIX"), self._zone_type

            if self._zone_type is None:
                self._set_zone_type("ELE")  # might eventually be: "VAL"

        elif msg.code == "000A":
            self._zone_config = msg

        elif msg.code == "12B0":
            self._window_open = msg

        elif msg.code == "2309" and msg.verb in (I_, RP):  # setpoint
            assert msg.src.type == "01", "coding error zxw"
            self._setpoint = msg

        elif msg.code == "2349" and msg.verb in (I_, RP):  # mode, setpoint
            assert msg.src.type == "01", "coding error zxx"
            self._mode = msg
            self._setpoint = msg

        elif msg.code == "30C9" and msg.verb in (I_, RP):  # used by sensor matching
            assert msg.src.type in DEVICE_HAS_ZONE_SENSOR + ("01",), "coding error"
            self._temperature = msg

        elif msg.code == "3150":  # TODO: and msg.verb in (I_, RP)?
            assert msg.src.type in ("00", "02", "04", "13")
            assert self._zone_type in (None, "RAD", "UFH", "VAL")  # MIX/ELE don't 3150

            if msg.src.type in ("00", "02", "04", "13"):
                zone_type = ZONE_CLASS_MAP[msg.src.type]
                self._set_zone_type("VAL" if zone_type == "ELE" else zone_type)

        # elif "zone_idx" in msg.payload:
        #     pass

    @property
    def sensor(self) -> Device:
        return self._sensor

    def _set_sensor(self, device: Device) -> None:  # self._sensor
        """Set the temp sensor for this zone (one of: 01:, 03:, 04:, 12:, 22:, 34:)."""

        if self._sensor is device:
            return
        if self._sensor is not None:
            raise CorruptStateError(
                f"{self} changed {ATTR_ZONE_SENSOR}: {self._sensor} to {device}"
            )

        sensor_types = ("00", "01", "03", "04", "12", "22", "34")
        if not isinstance(device, Device) or device.type not in sensor_types:
            # TODO: or not hasattr(device, "temperature")
            raise TypeError(f"{self}: {ATTR_ZONE_SENSOR} can't be: {device}")

        self._sensor = device
        device._set_parent(self)  # , domain=self.idx)

    @property
    def heating_type(self) -> Optional[str]:
        """TODO.

        There are three ways to determine the type of a zone:
        1. Use a 0005 packet (deterministic)
        2. Eavesdrop (non-deterministic, slow to converge)
        3. via a config file (a schema)
        """

        if self._zone_type is not None:  # isinstance(self, ???)
            return ZONE_TYPE_MAP.get(self._zone_type)

        # TODO: actuators
        dev_types = [d.type for d in self.devices if d.type in ("00", "02", "04", "13")]

        if "02" in dev_types:
            zone_type = "UFH"
        elif "13" in dev_types:
            zone_type = "VAL" if "3150" in self._msgs else "ELE"
        # elif "??" in dev_types:  # TODO:
        #     zone_type = "MIX"
        elif "04" in dev_types or "00" in dev_types:
            # beware edge case: TRV as sensor for a non-RAD zone
            zone_type = "RAD"
        else:
            zone_type = None

        if zone_type is not None:
            self._set_zone_type(zone_type)

        return ZONE_TYPE_MAP.get(self._zone_type)

    def _set_zone_type(self, zone_type: str):  # self._zone_type
        """Set the zone's type, after validating it.

        There are two possible sources for the type of a zone:
        1. eavesdropping packet codes
        2. analyzing child devices

        Both will execute a zone.type = type (i.e. via this setter).
        """

        _type = ZONE_TYPE_SLUGS.get(zone_type, zone_type)
        if _type not in ZONE_BY_CLASS_ID:
            raise ValueError(f"Not a known zone_type: {zone_type}")

        if self._zone_type == _type:
            return
        if self._zone_type is not None and (
            self._zone_type != "ELE" and _type != "VAL"
        ):
            raise CorruptStateError(
                f"{self} changed zone_type: {self._zone_type} to {_type}"
            )

        self._zone_type = _type
        self.__class__ = ZONE_BY_CLASS_ID[_type]
        self._discover()  # TODO: needs tidyup (ref #67)

    @property
    def name(self) -> Optional[str]:
        """Return the name of the zone."""
        return self._msg_payload(self._name, "name")

    @name.setter
    def name(self, value) -> Optional[str]:
        """Set the name of the zone."""
        self._gwy.send_cmd(Command.set_zone_name(self._ctl.id, self.idx, value))

        # async def get_name(self, force_refresh=None) -> Optional[str]:
        #     """Return the name of the zone."""
        #     if not force_refresh and self._name is not None:
        #         return self._msg_payload(self._name, "name")

        #     self._name = None
        #     self._send_cmd("0004", payload=f"{self.idx}00")
        #     while self._name is None:
        #         await asyncio.sleep(0.05)

        #     return self._msg_payload(self._name, "name")

    @property
    def config(self) -> Optional[dict]:  # 000A
        if not self._zone_config or self._zone_config.is_expired:
            return

        elif isinstance(self._zone_config.payload, dict):
            result = self._msg_payload(self._zone_config)

        elif isinstance(self._zone_config.payload, list):
            tmp = [z for z in self._zone_config.payload if z["zone_idx"] == self.idx]
            result = {k: v for k, v in tmp[0].items() if k[:1] != "_"}

        if result:
            return {k: v for k, v in result.items() if k != "zone_idx"}

    @property
    def mode(self) -> Optional[dict]:  # 2349
        # setpoint, mode, until
        if not self._mode or self._mode.is_expired:
            return

        result = self._msg_payload(self._mode)
        if result:
            return {k: v for k, v in result.items() if k != "zone_idx"}

    @property
    def setpoint(self) -> Optional[float]:  # 2309 (2349 is a superset of 2309)
        if not self._setpoint or self._setpoint.is_expired:
            return

        elif isinstance(self._setpoint.payload, dict):
            return self._msg_payload(self._setpoint, ATTR_SETPOINT)

        elif isinstance(self._setpoint.payload, list):
            _zone = [z for z in self._setpoint.payload if z["zone_idx"] == self.idx]
            return _zone[0][ATTR_SETPOINT]

    @setpoint.setter
    def setpoint(self, value) -> None:
        """Set the target temperature, until the next scheduled setpoint."""
        if value is None:
            self.reset_mode()
        else:
            cmd = Command.set_zone_setpoint(self._ctl.id, self.idx, value)
            self._gwy.send_cmd(cmd)
            # NOTE: the following doesn't wotk for e.g. Hometronics
            # self.set_mode(mode=ZoneMode.ADVANCED, setpoint=value)

    @property
    def temperature(self) -> Optional[float]:  # 30C9
        # if not self._temperature or self._temperature.is_expired:
        #       # NOTE: CTL (as sensor) won't have this attr...
        #     if self.sensor and self.sensor.temperature:
        #         self._temperature = self.sensor._temp

        if not self._temperature or (
            self._temperature.is_expired == self._temperature.HAS_EXPIRED
        ):
            return

        elif isinstance(self._temperature.payload, dict):
            return self._msg_payload(self._temperature, ATTR_TEMP)

        elif isinstance(self._temperature.payload, list):
            tmp = [z for z in self._temperature.payload if z["zone_idx"] == self.idx]
            return tmp[0][ATTR_TEMP]

    @property
    def heat_demand(self) -> Optional[float]:  # 3150
        """Return the zone's heat demand, estimated from its devices' heat demand."""
        demands = [
            d.heat_demand
            for d in self.devices
            if hasattr(d, ATTR_HEAT_DEMAND) and d.heat_demand is not None
        ]
        # return round(sum(demands) / len(demands), 1) if demands else None
        return max(demands + [0]) if demands else None

    @property
    def window_open(self) -> Optional[bool]:  # 12B0  # TODO: don't work >1 TRV?
        """Return an estimate of the zone's current window_open state."""
        windows = [
            d.window_open
            for d in self.devices
            if hasattr(d, ATTR_WINDOW_OPEN) and d.window_open is not None
        ]
        return any(windows) if windows else None

    def reset_config(self) -> Task:
        """Reset the zone's parameters to their default values."""
        return self.set_config()

    def set_config(
        self,
        min_temp=5,
        max_temp=35,
        local_override: bool = False,
        openwindow_function: bool = False,
        multiroom_mode: bool = False,
    ) -> Task:
        """Set the zone's parameters (min_temp, max_temp, etc.)."""
        cmd = Command.set_zone_config(
            self._ctl.id,
            self.idx,
            min_temp=min_temp,
            max_temp=max_temp,
            local_override=local_override,
            openwindow_function=openwindow_function,
            multiroom_mode=multiroom_mode,
        )
        return self._gwy.send_cmd(cmd)

    def reset_mode(self) -> Task:  # 2349
        """Revert the zone to following its schedule."""
        return self.set_mode(mode=ZoneMode.SCHEDULE)

    def set_frost_mode(self) -> Task:  # 2349
        """Set the zone to the lowest possible setpoint, indefinitely."""
        return self.set_mode(mode=ZoneMode.PERMANENT, setpoint=5)  # TODO

    def set_mode(self, mode=None, setpoint=None, until=None) -> Task:
        """Override the zone's setpoint for a specified duration, or indefinitely."""
        cmd = Command.set_zone_mode(self._ctl.id, self.idx, mode, setpoint, until)
        return self._gwy.send_cmd(cmd)

    def set_name(self, name) -> Task:
        """Set the zone's name."""
        cmd = Command.set_zone_name(self._ctl.id, self.idx, name)
        return self._gwy.send_cmd(cmd)

    @property
    def schema(self) -> dict:
        """Return the zone's schema (type, devices)."""
        if not self._sensor:
            sensor_schema = None
        elif getattr(self._sensor, "_30C9_faked", None) is None:
            # NOTE: CTL (as sensor) won't have this attr...
            sensor_schema = self._sensor.id
        else:
            sensor_schema = {
                "device_id": self._sensor.id,
                "is_faked": self._sensor._30C9_faked,
            }

        return {
            ATTR_ZONE_TYPE: self.heating_type,
            ATTR_ZONE_SENSOR: sensor_schema,
            ATTR_DEVICES: [d.id for d in self.devices],
        }

    @property  # TODO: setpoint
    def params(self) -> dict:
        """Return the zone's configuration (excl. schedule)."""
        return {a: getattr(self, a) for a in ("config", "mode", "name")}

    @property
    def status(self) -> dict:
        """Return the zone's current state."""
        return {
            a: getattr(self, a) for a in (ATTR_SETPOINT, ATTR_TEMP, ATTR_HEAT_DEMAND)
        }


class EleZone(RelayDemand, Zone):  # BDR91A/T  # TODO: 0008/0009/3150
    """For a small electric load controlled by a relay (never calls for heat)."""

    __zon_class__ = ZONE_CLASS.ELE

    # def __init__(self, *args, **kwargs) -> None:  # can't use this here

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        # NOTE: we create, then promote, so shouldn't super()
        # super()._discover(discover_flag=discover_flag)
        if False and discover_flag & DISCOVER_SCHEMA:
            self._send_cmd("000C", payload=f"{self.idx}{_000C_DEVICE.ELE}")

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        # if msg.code == "0008":  # ZON zones are ELE zones that also call for heat
        #     self._set_zone_type("VAL")
        if msg.code == "3150":
            raise TypeError("WHAT 1")
        elif msg.code == "3EF0":
            raise TypeError("WHAT 2")

    @property
    def heat_demand(self) -> None:  # Electric zones (do *not* call for heat)
        """Return None as the zone's heat demand, electric zones don't call for heat."""
        return


class MixZone(Zone):  # HM80  # TODO: 0008/0009/3150
    """For a modulating valve controlled by a HM80 (will also call for heat).

    Note that HM80s are listen-only devices.
    """

    __zon_class__ = ZONE_CLASS.MIX

    # def __init__(self, *args, **kwargs) -> None:  # can't use this here

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        # NOTE: we create, then promote, so shouldn't super()
        # super()._discover(discover_flag=discover_flag)
        if False and discover_flag & DISCOVER_SCHEMA:
            self._send_cmd("000C", payload=f"{self.idx}{_000C_DEVICE.MIX}")

        if discover_flag & DISCOVER_PARAMS:
            self._gwy.send_cmd(Command.get_mix_valve_params(self._ctl.id, self.idx))

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        if msg.code == "1030" and msg.verb == I_:
            self._mix_config = msg

    @property
    def mix_config(self) -> dict:
        return self._msg_payload(self._mix_config)

    @property
    def params(self) -> dict:
        return {**super().status, "mix_config": self.mix_config}


class RadZone(Zone):  # HR92/HR80
    """For radiators controlled by HR92s or HR80s (will also call for heat)."""

    __zon_class__ = ZONE_CLASS.RAD

    # def __init__(self, *args, **kwargs) -> None:  # can't use this here

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        # NOTE: we create, then promote, so shouldn't super()
        # super()._discover(discover_flag=discover_flag)
        if False and discover_flag & DISCOVER_SCHEMA:
            self._send_cmd("000C", payload=f"{self.idx}{_000C_DEVICE.RAD}")


class UfhZone(Zone):  # HCC80/HCE80  # TODO: needs checking
    """For underfloor heating controlled by an HCE80/HCC80 (will also call for heat)."""

    __zon_class__ = ZONE_CLASS.UFH

    # def __init__(self, *args, **kwargs) -> None:  # can't use this here

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        # NOTE: we create, then promote, so shouldn't super()
        # super()._discover(discover_flag=discover_flag)
        if False and discover_flag & DISCOVER_SCHEMA:
            self._send_cmd("000C", payload=f"{self.idx}{_000C_DEVICE.UFH}")

    def _handle_msg(self, msg) -> bool:
        super()._handle_msg(msg)

        if msg.code == "22C9" and msg.verb == I_:
            self._ufh_setpoint = msg

    @property
    def ufh_setpoint(self) -> Optional[float]:  # 22C9
        return self._msg_payload(self._ufh_setpoint, ATTR_SETPOINT)

    @property
    def status(self) -> dict:
        return {**super().status, "ufh_setpoint": self.ufh_setpoint}


class ValZone(EleZone):  # BDR91A/T
    """For a motorised valve controlled by a BDR91 (will also call for heat)."""

    __zon_class__ = ZONE_CLASS.VAL

    # def __init__(self, *args, **kwargs) -> None:  # can't use this here

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        # NOTE: we create, then promote, so shouldn't super()
        # super()._discover(discover_flag=discover_flag)
        if False and discover_flag & DISCOVER_SCHEMA:
            self._send_cmd("000C", payload=f"{self.idx}{_000C_DEVICE.VAL}")

    @property
    def heat_demand(self) -> Optional[float]:  # 0008 (NOTE: not 3150)
        """Return the zone's heat demand, using relay demand as a proxy."""
        return self.relay_demand


CLASS_ATTR = "__zon_class__"
ZONE_BY_CLASS_ID = {
    getattr(c[1], CLASS_ATTR): c[1]
    for c in getmembers(
        modules[__name__],
        lambda m: isclass(m) and m.__module__ == __name__ and hasattr(m, CLASS_ATTR),
    )
}  # e.g. "RAD": RadZone


def create_zone(evo, zone_idx, profile=None, **kwargs) -> Zone:
    """Create a zone, and optionally perform discovery & start polling."""

    if profile is None:
        profile = ZONE_CLASS.DHW if zone_idx == "HW" else None

    zone = ZONE_BY_CLASS_ID.get(profile, Zone)(evo, zone_idx, **kwargs)

    if not evo._gwy.config.disable_discovery:
        schedule_task(zone._discover, discover_flag=DISCOVER_SCHEMA, delay=1)
        schedule_task(zone._discover, discover_flag=DISCOVER_PARAMS, delay=4)
        schedule_task(zone._discover, discover_flag=DISCOVER_STATUS, delay=7)

    return zone



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

Decode/process a message (payload into JSON).
"""

import logging
import re
from datetime import datetime as dt
from datetime import timedelta as td
from typing import Any, Optional, Tuple, Union

from . import parsers
from .const import (
    _0005_ZONE_TYPE,
    ATTR_DHW_SENSOR,
    ATTR_DHW_VALVE,
    ATTR_DHW_VALVE_HTG,
    ATTR_HTG_CONTROL,
    ATTR_ZONE_ACTUATORS,
    ATTR_ZONE_SENSOR,
    DEVICE_TYPES,
    MSG_FORMAT_10,
    MSG_FORMAT_18,
    NON_DEV_ADDR,
    NUL_DEV_ADDR,
    ZONE_TYPE_SLUGS,
    Address,
    __dev_mode__,
)
from .devices import Device, FanDevice, OtbGateway
from .exceptions import (
    CorruptEvohomeError,
    CorruptPacketError,
    CorruptPayloadError,
    CorruptStateError,
)
from .opentherm import MSG_ID
from .packet import _PKT_LOGGER  # TODO: bad packets are being logged twice!
from .ramses import CODES_WITH_COMPLEX_IDX, CODES_WITHOUT_IDX, RAMSES_CODES
from .schema import DONT_CREATE_ENTITIES, DONT_UPDATE_ENTITIES

# from .systems import Evohome

CODE_NAMES = {k: v["name"] for k, v in RAMSES_CODES.items()}

I_, RQ, RP, W_ = " I", "RQ", "RP", " W"

# TODO: WIP
MSG_TIMEOUTS = {
    "0004": {I_: td(days=1), RP: td(days=1)},
    "000A": {I_: td(days=1), RP: td(days=1)},
    "0100": {I_: td(days=1), RP: td(days=1)},
    "1060": {I_: td(days=1)},
    "10A0": {RP: td(hours=4)},
    "1100": {I_: td(days=1), RP: td(days=1)},
    "1260": {I_: td(hours=1), RP: td(hours=1)},
    "12B0": {I_: td(hours=1), RP: td(hours=1)},
    "1F41": {I_: td(hours=4), RP: td(hours=4)},
    "2309": {I_: td(minutes=30), RP: td(minutes=30)},
    "2349": {I_: td(hours=4), RP: td(hours=4)},
    "2E04": {I_: td(hours=4), RP: td(hours=4)},
    "313F": {I_: td(seconds=3), RP: td(seconds=3)},
    "3150": {I_: td(minutes=20)},
    "3C09": {I_: td(hours=4), RP: td(hours=4)},
}  # not arrays

DEV_MODE = True or __dev_mode__ and False

_LOGGER = logging.getLogger(__name__)
if DEV_MODE:
    _LOGGER.setLevel(logging.DEBUG)

# _PKT_LOGGER = _LOGGER


class Message:
    """The message class.

    Will trap/log all invalid msgs appropriately.
    """

    HAS_EXPIRED = 2
    IS_EXPIRING = 1
    NOT_EXPIRED = 0

    def __init__(self, gwy, pkt) -> None:
        """Create a message, assumes a valid packet."""
        self._gwy = gwy
        self._pkt = pkt

        # prefer Devices but should use Addresses for now...
        self.src = pkt.src_addr  # gwy.device_by_id.get(pkt.src_addr.id, pkt.src_addr)
        self.dst = pkt.dst_addr  # gwy.device_by_id.get(pkt.dst_addr.id, pkt.dst_addr)

        self.devs = pkt.addrs
        self.date = pkt.date
        self.time = pkt.time
        self.dtm = pkt._dtm

        self.rssi = pkt.packet[0:3]
        self.verb = pkt.packet[4:6]
        self.seqn = pkt.packet[7:10]
        self.code = pkt.packet[41:45]
        self.len = int(pkt.packet[46:49])
        self.raw_payload = pkt.packet[50:]

        self.code_name = CODE_NAMES.get(self.code, f"unknown_{self.code}")
        self._payload = None
        self._str = None

        self._haz_payload = None
        self._haz_simple_idx = None

        self._is_array = None
        self._is_expired = None
        self._is_fragment = None

        self._is_valid = None
        if not self.is_valid:
            raise ValueError("not a valid message")

    def __repr__(self) -> str:
        """Return an unambiguous string representation of this object."""
        return self._pkt.packet

    def __str__(self) -> str:
        """Return a brief readable string representation of this object."""

        def display_name(dev: Union[Address, Device]) -> str:
            """Return a formatted device name, uses a friendly name if there is one."""
            if self._gwy.config.use_names:
                try:
                    return f"{self._gwy.known_devices[dev.id]['name']:<18}"
                except (KeyError, TypeError):
                    pass

            if dev is NON_DEV_ADDR:
                return f"{'':<10}"

            if dev is NUL_DEV_ADDR:
                return "NUL:------"

            return f"{DEVICE_TYPES.get(dev.type, f'{dev.type:>3}')}:{dev.id[3:]}"
            try:
                return f"{DEVICE_TYPES.get(dev.type, f'{dev.type:>3}')}:{dev.id[3:]}"
            except AttributeError:  # TODO: 'NoneType' object has no attribute 'type'
                return "XXX:------"

        if self._str is not None:
            return self._str

        if not self.is_valid:
            return  # "Invalid"

        if self.src.id == self.devs[0].id:
            src = display_name(self.src)
            dst = display_name(self.dst) if self.dst is not self.src else ""
        else:
            src = ""
            dst = display_name(self.src)

        payload = self.raw_payload if self.len < 4 else f"{self.raw_payload[:5]}..."[:9]

        _format = MSG_FORMAT_18 if self._gwy.config.use_names else MSG_FORMAT_10
        self._str = _format.format(
            src, dst, self.verb, self.code_name, payload, self._payload
        )
        return self._str

    def __eq__(self, other) -> bool:
        if not isinstance(other, Message):
            return NotImplemented
        return all(
            self.verb == other.verb,
            self.code == other.code,
            self.src.id == other.src.id,
            self.dst.id == other.dst.id,
            self.raw_payload == other.raw_payload,
        )

    def __lt__(self, other) -> bool:
        if not isinstance(other, Message):
            return NotImplemented
        return self.dtm < other.dtm

    @property
    def _has_payload(self) -> bool:
        """Return False if no payload (may falsely Return True).

        The message (i.e. the raw payload) may still have an idx.
        """

        if self._haz_payload is not None:
            return self._haz_payload

        if self.len == 1:
            self._haz_payload = False  # TODO: (WIP) has no payload
        elif RAMSES_CODES.get(self.code):
            self._haz_payload = RAMSES_CODES[self.code].get(self.verb) not in (
                r"^00$",
                r"^00(00)?$",
                r"^FF$",
            )
            assert (
                self._haz_payload or self.len < 3
            ), "Message not expected to have a payload! Is it corrupt?"
        else:
            self._haz_payload = True

        return self._haz_payload

    @property
    def _has_simple_idx(self) -> bool:
        """Return False if no index, or if it is complex.(may falsely Return True).

        The message may still have a payload.
        """

        if self._haz_simple_idx is not None:
            return self._haz_simple_idx

        if self.code in CODES_WITH_COMPLEX_IDX:
            self._haz_simple_idx = False  # has (complex) _idx via their parser
        elif self.code in CODES_WITHOUT_IDX:
            self._haz_simple_idx = False  # has no idx, even though some != "00"
        elif RAMSES_CODES.get(self.code):
            self._haz_simple_idx = (
                RAMSES_CODES[self.code].get(self.verb, "")[:3] != "^00"
            )  # has no _idx
            assert (
                self._haz_simple_idx or self.raw_payload[:2] == "00"
            ), "Message not expected to have a index! Is it corrupt?"
        else:
            self._haz_simple_idx = True

        return self._haz_simple_idx

    @property
    def payload(self) -> Any:  # Any[dict, List[dict]]:
        """Return the payload."""
        if self._payload is not None:
            return self._payload

        if self.is_valid:
            if self.code != "000C":  # TODO: assert here, or in is_valid()
                assert self.is_array == isinstance(self._payload, list)
            return self._payload

    @property
    def is_array(self) -> bool:
        """Return True if the message's raw payload is an array.

        Note that the corresponding parsed payload may not match, e.g. the 000C payload
        is not a list.

        Does not require a valid payload.
        """

        if self._is_array is not None:
            return self._is_array

        if self.verb in (RQ, W_):
            self._is_array = False

        elif self.code in ("000C", "1FC9"):  # also: 0005?
            # grep -E ' (I|RP).* 000C '  #  from 01:/30: (VMS) only
            # grep -E ' (I|RP).* 1FC9 '  #  from 01:/13:/other (not W)
            self._is_array = True

        elif self.verb == RP:
            self._is_array = False

        # 087  I 092 --:------ --:------ 12:126457 2309 006 0107D0-020708
        # 090  I 093 --:------ --:------ 12:126457 30C9 003 017FFF
        # 089  I 094 --:------ --:------ 12:126457 000A 012 010001F40BB8-020001F40BB8
        elif self.code in ("000A", "2309", "30C9") and self.src.type == "12":
            self._is_array = self.verb == I_ and self.dst.id == "--:------"

        elif self.verb not in (I_, RP) or self.src.id != self.dst.id:
            self._is_array = False

        # 045  I --- 01:158182 --:------ 01:158182 0009 003 0B00FF (or: FC00FF)
        # 045  I --- 01:145038 --:------ 01:145038 0009 006 FC00FFF900FF
        elif self.code in ("0009",) and self.src.type == "01":
            # grep -E ' I.* 01:.* 01:.* 0009 [0-9]{3} F' (and: grep -v ' 003 ')
            self._is_array = self.verb == I_ and self.raw_payload[:1] == "F"

        elif self.code in ("000A", "2309", "30C9") and self.src.type == "01":
            # grep ' I.* 01:.* 01:.* 000A '
            # grep ' I.* 01:.* 01:.* 2309 ' | grep -v ' 003 '  # TODO: some non-arrays
            # grep ' I.* 01:.* 01:.* 30C9 '
            self._is_array = self.verb == I_ and self.src.id == self.dst.id

        # 055  I --- 02:001107 --:------ 02:001107 22C9 024 0008340A28010108340A...
        # 055  I --- 02:001107 --:------ 02:001107 22C9 006 0408340A2801
        # 055  I --- 02:001107 --:------ 02:001107 3150 010 00640164026403580458
        # 055  I --- 02:001107 --:------ 02:001107 3150 010 00000100020003000400
        elif self.code in ("22C9", "3150") and self.src.type == "02":
            # grep -E ' I.* 02:.* 02:.* 22C9 '
            # grep -E ' I.* 02:.* 02:.* 3150' | grep -v FC
            self._is_array = self.verb == I_ and self.src.id == self.dst.id
            self._is_array = self._is_array if self.raw_payload[:1] != "F" else False

        # 095  I --- 23:100224 --:------ 23:100224 2249 007 007EFF7EFFFFFF
        # 095  I --- 23:100224 --:------ 23:100224 2249 007 007EFF7EFFFFFF
        elif self.code in ("2249",) and self.src.type == "23":
            self._is_array = self.verb == I_ and self.src.id == self.dst.id
            # self._is_array = self._is_array if self.raw_payload[:1] != "F" else False

        else:
            self._is_array = False

        return self._is_array

    @property
    def is_expired(self) -> Tuple[bool, Optional[bool]]:
        """Return True if the message is dated (does not require a valid payload)."""

        def _logger_send(logger, message) -> None:
            if True or DEV_MODE:  # TODO: remove 'True or'
                logger(
                    "Message(%s), received at %s: %s",
                    self._pkt._header,
                    self.dtm,
                    message,
                )

        def _timeout() -> td:  # TODO: move this to RAMSES pkt schema
            timeout = None

            if self.verb in (RQ, W_):
                timeout = td(seconds=3)

            elif self.code in ("0005", "000C", "10E0"):
                return  # TODO: exclude/remove devices caused by corrupt ADDRs?

            elif self.code == "1FC9" and self.verb == RP:
                return  # TODO: check other verbs, they seem variable

            elif self.code == "1F09":
                timeout = td(seconds=self.payload["remaining_seconds"])

            elif self.code == "000A" and self._is_array:
                timeout = td(minutes=60)  # sends I /1h

            elif self.code in ("2309", "30C9") and self._is_array:
                timeout = td(minutes=15)  # sends I /sync_cycle

            elif self.code == "3220":
                if self.payload[MSG_ID] in OtbGateway.SCHEMA_MDG_IDS:
                    timeout = None
                elif self.payload[MSG_ID] in OtbGateway.PARAMS_MDG_IDS:
                    timeout = td(minutes=60)
                else:  # incl. OtbGateway.STATUS_MDG_IDS
                    timeout = td(minutes=5)

            elif self.code in MSG_TIMEOUTS and self.verb in MSG_TIMEOUTS[self.code]:
                return MSG_TIMEOUTS[self.code][self.verb]

            # elif self.code in ("3B00", "3EF0", ):  # TODO: 0008, 3EF0, 3EF1
            #     timeout = td(minutes=6.7)  # TODO: WIP

            return timeout or td(minutes=60)

        if self._is_expired == self.HAS_EXPIRED:  # TODO: or CANT_EXPIRE
            return self._is_expired

        timeout = _timeout()  # TODO: needs fixing

        if timeout is None:  # treat as never expiring
            self._is_expired = self.NOT_EXPIRED  # TODO: CANT_EXPIRE
            # _logger_send(
            #    _LOGGER.debug,
            #    "msg is not expirable",
            # )
            return self._is_expired

        dtm_now = (
            dt.now()
            if self._gwy.serial_port
            else (self._gwy._prev_msg.dtm if self._gwy._prev_msg else self.dtm)
        )

        if self.dtm < dtm_now - timeout * 2:
            if self._is_expired != self.HAS_EXPIRED:
                _logger_send(
                    _LOGGER.error,
                    f"msg has tombstoned ({dtm_now - self.dtm}, {timeout})",
                )
            self._is_expired = self.HAS_EXPIRED

        elif self.dtm < dtm_now - timeout * 1:
            if self._is_expired != self.IS_EXPIRING:
                _logger_send(
                    _LOGGER.info,
                    f"msg has expired ({dtm_now - self.dtm}, {timeout})",
                )
            self._is_expired = self.IS_EXPIRING

        else:
            self._is_expired = self.NOT_EXPIRED
            # _logger_send(
            #     _LOGGER.debug,
            #     f"msg has not expired ({dtm_now - self.dtm}, {timeout})",
            # )
        return self._is_expired

    @property
    def _is_fragment_WIP(self) -> bool:
        """Return True if the raw payload is a fragment of a message."""

        # if not self._is_valid:
        #     return
        if self._is_fragment is not None:
            return self._is_fragment

        # packets have a maximum length of 48 (decimal)
        # if self.code == "000A" and self.verb == I_:
        #     self._is_fragment = True if len(???.zones) > 8 else None
        # el
        if self.code == "0404" and self.verb == RP:
            self._is_fragment = True
        elif self.code == "22C9" and self.verb == I_:
            self._is_fragment = None  # max length 24!
        else:
            self._is_fragment = False

        return self._is_fragment

    @property
    def is_valid(self) -> bool:  # Main code here
        """Parse the payload, return True if the message payload is valid.

        All exceptions are trapped, and logged appropriately.
        """

        def log_message(log_level, log_msg):
            log_level(log_msg, self._pkt, extra=self._pkt.__dict__)

        if self._is_valid is not None:
            return self._is_valid
        self._is_valid = False  # Assume is invalid

        try:  # determine which parser to use
            payload_parser = getattr(parsers, f"parser_{self.code}".lower())

        except AttributeError:  # there's no parser for this command code!
            payload_parser = getattr(parsers, "parser_unknown")

        try:  # parse the packet
            self._payload = payload_parser(self.raw_payload, self)
            assert isinstance(self._payload, dict) or isinstance(
                self._payload, list
            ), "message payload is not dict nor list"

        except AssertionError as err:
            # beware: HGI80 can send parseable but 'odd' packets +/- get invalid reply
            hint = f": {err}" if str(err) != "" else ""
            if not hint or DEV_MODE:
                log_message(_PKT_LOGGER.exception, "%s < Validation error ")
            elif self.src.type != "18":
                log_message(_PKT_LOGGER.exception, f"%s < Validation error{hint} ")
            else:  # elif DEV_MODE:  # TODO: consider info/debug for the following
                log_message(_PKT_LOGGER.exception, f"%s < Validation error{hint} ")

        except (CorruptPacketError, CorruptPayloadError) as err:  # CorruptEvohomeError
            if DEV_MODE:
                log_message(_PKT_LOGGER.exception, f"%s < {err}")
            else:
                log_message(_PKT_LOGGER.warning, f"%s < {err}")

        except (AttributeError, LookupError, TypeError, ValueError):  # TODO: dev only
            log_message(_PKT_LOGGER.exception, "%s < Coding error ")

        except NotImplementedError:  # parser_unknown (unknown packet code)
            log_message(_PKT_LOGGER.warning, "%s < Unknown packet code ")

        else:
            self._is_valid = True

        return self._is_valid


def process_msg(msg: Message) -> None:
    """Decode the packet and its payload.

    All methods require a valid message (payload), except create_devices, which requires
    a valid message only for 000C.
    """

    def create_devices(this) -> None:
        """Discover and create any new devices."""

        if this.code == "000C" and this.verb == RP:
            if this.src.type == "01":  # TODO
                this._gwy._get_device(this.dst, ctl_addr=this.src)
                if this.is_valid:
                    key = "zone_idx" if "zone_idx" in this.payload else "domain_id"
                    [
                        this._gwy._get_device(
                            Address(id=d, type=d[:2]),
                            ctl_addr=this.src,
                            domain_id=this.payload[key],
                        )
                        for d in this.payload["devices"]
                    ]
            if this.src.type == "02":  # TODO
                # this._gwy._get_device(this.dst)
                if this.payload["devices"]:
                    device_id = this.payload["devices"][0]
                    this._gwy._get_device(
                        this.src, ctl_addr=Address(id=device_id, type=device_id[:2])
                    )

        elif this.code in ("31D9", "31DA", "31E0") and this.verb in (I_, RP):
            device = this._gwy._get_device(this.src)
            if device.__class__ is Device:
                device.__class__ = FanDevice  # HACK: because my HVAC is a 30:

        if this.src.type in ("01", "23") and this.src is not this.dst:  # TODO: all CTLs
            this.src = this._gwy._get_device(this.src, ctl_addr=this.src)
            ctl_addr = this.src if msg._gwy.config.enable_eavesdrop else None
            this._gwy._get_device(this.dst, ctl_addr=ctl_addr)

        elif this.dst.type in ("01", "23") and this.src is not this.dst:  # all CTLs
            this.dst = this._gwy._get_device(this.dst, ctl_addr=this.dst)
            ctl_addr = this.dst if msg._gwy.config.enable_eavesdrop else None
            this._gwy._get_device(this.src, ctl_addr=ctl_addr)

        # TODO: will need other changes before these two will work...
        # TODO: the issue is, if the 1st pkt is not a 1F09 (or a list 000A/2309/30C9)
        # TODO: also could do 22D9 (UFC), others?
        # elif this.code == "1F09" and this.verb == I_:
        #     this._gwy._get_device(this.dst, ctl_addr=this.src)

        # TODO: ...such as means to promote a device to a controller

        # this should catch all non-controller (and *some* controller) devices
        elif this.src is this.dst:
            this._gwy._get_device(this.src)

        # otherwise one will be a controller, *unless* dst is in ("--", "63")
        elif isinstance(this.src, Device) and this.src._is_controller:
            this._gwy._get_device(this.dst, ctl_addr=this.src)

        # TODO: may create a controller that doesn't exist
        elif isinstance(this.dst, Device) and this.dst._is_controller:
            this._gwy._get_device(this.src, ctl_addr=this.dst)

        else:
            # beware:  I --- --:------ --:------ 10:078099 1FD4 003 00F079
            [this._gwy._get_device(d) for d in (this.src, this.dst)]

        # where possible, swap each Address for its corresponding Device
        this.src = this._gwy.device_by_id.get(this.src.id, this.src)
        if this.dst is not None:
            this.dst = this._gwy.device_by_id.get(this.dst.id, this.dst)

    def create_zones(this) -> None:
        """Discover and create any new zones (except HW)."""

        if this.src.type not in ("01", "23"):  # TODO: this is too restrictive!
            return

        evo = this.src._evo

        # TODO: a I/0005: zones have changed & may need a restart (del) or not (add)
        if this.code == "0005":  # RP, and also I
            if this._payload["zone_type"] in _0005_ZONE_TYPE.values():
                [
                    evo._get_zone(
                        f"{idx:02X}",
                        zone_type=ZONE_TYPE_SLUGS.get(this._payload["zone_type"]),
                    )
                    for idx, flag in enumerate(this._payload["zone_mask"])
                    if flag == 1
                ]

        if this.code == "000C" and this.src.type == "01":
            if this.payload["devices"]:  # i.e. if len(devices) > 0
                devices = [this.src.device_by_id[d] for d in this.payload["devices"]]

                if this.payload["device_class"] == ATTR_ZONE_SENSOR:
                    zone = evo._get_zone(this.payload["zone_idx"])
                    try:
                        zone._set_sensor(devices[0])
                    except TypeError:  # ignore invalid device types, e.g. 17:
                        pass

                elif this.payload["device_class"] == ATTR_ZONE_ACTUATORS:
                    # evo._get_zone(this.payload["zone_idx"], actuators=devices)
                    # TODO: whihc is better, above or below?
                    zone = evo._get_zone(this.payload["zone_idx"])
                    [d._set_parent(zone) for d in devices]

                elif this.payload["device_class"] == ATTR_HTG_CONTROL:
                    evo._set_htg_control(devices[0])

                elif this.payload["device_class"] == ATTR_DHW_SENSOR:
                    evo._get_dhw()._set_sensor(devices[0])

                elif this.payload["device_class"] == ATTR_DHW_VALVE:
                    evo._get_dhw()._set_dhw_valve(devices[0])

                elif this.payload["device_class"] == ATTR_DHW_VALVE_HTG:
                    evo._get_dhw()._set_htg_valve(devices[0])

            elif this.payload["device_class"] == ATTR_HTG_CONTROL:
                # TODO: maybe the htg controller is an OTB? via eavesdropping
                # evo._set_htg_control(devices[0])
                pass

        # # Eavesdropping (below) is used when discovery (above) is not an option
        # # TODO: needs work, e.g. RP/1F41 (excl. null_rp)
        # elif this.code in ("10A0", "1F41"):
        #     if isinstance(this.dst, Device) and this.dst._is_controller:
        #         this.dst._get_dhw()
        #     else:
        #         evo._get_dhw()

        # # TODO: also process ufh_idx (but never domain_id)
        # elif isinstance(this._payload, dict):
        #     # TODO: only creating zones from arrays, presently, but could do so here
        #     if this._payload.get("zone_idx"):  # TODO: parent_zone too?
        #         if this.src._is_controller:
        #             evo._get_zone(this._payload["zone_idx"])
        #         else:
        #             this.dst._get_zone(this._payload["zone_idx"])

        elif isinstance(this._payload, list):
            if this.code in ("000A", "2309", "30C9"):  # the sync_cycle pkts
                [evo._get_zone(d["zone_idx"]) for d in this.payload]
            # elif this.code in ("22C9", "3150"):  # TODO: UFH zone
            #     pass

        # else:  # should never get here
        #     raise TypeError

    def update_entities(this, prev) -> None:  # TODO: needs work
        """Update the state of entities (devices, zones, ufh_zones)."""

        # HACK: merge 000A fragments
        # TODO: ?move to ctl._handle_msg() and/or system._handle_msg()?
        if re.search("I.* 01.* 000A ", str(this._pkt)):  # HACK: and dtm < 3 secs
            # TODO: an edge case here: >2 000A packets in a row
            if prev is not None and re.search("I.* 01.* 000A ", str(prev._pkt)):
                this._payload = prev.payload + this.payload  # merge frags, and process

        # some devices aren't created if they're filtered out (in create_devices?)
        if this.src not in this._gwy.devices:
            assert False, "what!!"
            return
            # 0008: BDR/RP, ir CTL/I/domain_id = F9/FA
            # 10A0: CTL/RP/dhw_idx
            # 1260: RP from CTL, or eavesdrop sensor
            # 1F41: eavesdrop

        # some empty payloads may still be useful (e.g. Rx/3EF0/{}, RQ/3EF1/{})
        this._gwy.device_by_id[this.src.id]._handle_msg(this)
        # if payload is {} (empty dict; lists shouldn't ever be empty)
        # if not this.payload:
        #     return

        # # TODO: what follows is a mess - needs sorting out
        # evo = this._gwy.system_by_id.get(this.src.id)
        # if evo:
        #     evo._handle_msg(this)

        for evo in this._gwy.systems:
            # if this.src == evo:  # TODO: or this.src.id == evo.id?
            if this.code in ("10A0", "1260", "1F41") and evo._dhw is not None:
                evo._dhw._handle_msg(this)
            break

        evo = this.src._evo if hasattr(this.src, "_evo") else None
        if evo is None:
            evo = this.dst._evo if hasattr(this.dst, "_evo") else None
        if evo is None:
            return

        if this.src is evo._ctl:
            evo._handle_msg(msg)

        if not hasattr(evo, "zone_by_idx"):
            return

        if isinstance(this.payload, dict) and "zone_idx" in this.payload:
            # 089  I --- 02:000921 --:------ 01:191718 3150 002 0300  # NOTE: is valid
            if this.payload["zone_idx"] in evo.zone_by_idx:
                evo.zone_by_idx[this.payload["zone_idx"]]._handle_msg(this)

            # elif "ufh_idx" in this.payload:
            #     # ufc = this.src._ufc if hasattr(this.src, "_ufc") else None
            #     pass

        elif isinstance(this.payload, list) and "zone_idx" in this.payload[0]:
            for z in this.payload:
                if z["zone_idx"] in evo.zone_by_idx:
                    evo.zone_by_idx[z["zone_idx"]]._handle_msg(this)

    if _LOGGER.getEffectiveLevel() == logging.INFO:  # i.e. don't log for DEBUG
        _LOGGER.info(msg)

    # TODO: This will need to be removed for HGI80-impersonation
    # 18:/RQs are unreliable, although any corresponding RPs are often required
    if msg.src.type == "18":
        return

    if msg._gwy.config.reduce_processing >= DONT_CREATE_ENTITIES:
        return

    try:  # process the payload
        create_devices(msg)  # from pkt header & from msg payload (e.g. 000C)
        # if msg._evo:  # TODO:
        create_zones(msg)  # create zones & (TBD) ufh_zones too?

        if msg._gwy.config.reduce_processing < DONT_UPDATE_ENTITIES:
            update_entities(msg, msg._gwy._prev_msg)  # update the state database

    except (AssertionError, NotImplementedError) as err:
        (_LOGGER.exception if DEV_MODE else _LOGGER.error)(
            "%s < %s", msg._pkt, err.__class__.__name__
        )
        return  # NOTE: use raise only when debugging

    except (AttributeError, LookupError, TypeError, ValueError) as err:
        (_LOGGER.exception if DEV_MODE else _LOGGER.error)(
            "%s < %s", msg._pkt, err.__class__.__name__
        )
        return  # NOTE: use raise only when debugging

    # except CorruptPacketError as err:
    #     (_LOGGER.exception if DEV_MODE else _LOGGER.error)(
    #         "%s < %s", msg._pkt, err
    #     )
    #     return  # NOTE: use raise only when debugging

    except CorruptStateError as err:
        (_LOGGER.exception if DEV_MODE else _LOGGER.error)("%s < %s", msg._pkt, err)
        return  # TODO: bad pkt, or Schema

    except CorruptEvohomeError as err:
        (_LOGGER.exception if DEV_MODE else _LOGGER.error)("%s < %s", msg._pkt, err)
        raise

    msg._gwy._prev_msg = msg if msg.is_valid else None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

Decode/process a packet (packet that was received).
"""

import logging
import shutil
import sys
from datetime import datetime as dt
from typing import Optional, Tuple

from .command import _pkt_header, _pkt_header_idx
from .const import MESSAGE_REGEX, __dev_mode__
from .exceptions import CorruptAddrSetError
from .helpers import extract_addrs
from .schema import LOG_FILE_NAME, LOG_ROTATE_BYTES, LOG_ROTATE_COUNT

DEV_MODE = __dev_mode__  # or True

DEFAULT_FMT = "%(asctime)s.%(msecs)03d %(message)s"
DEFAULT_DATEFMT = "%H:%M:%S"

try:
    import colorlog
except ModuleNotFoundError:
    _use_color_ = False
else:
    _use_color_ = True
    # basicConfig must be called after importing colorlog to ensure its handlers wrap
    # the correct streams
    # logging.basicConfig(format=DEFAULT_FMT, datefmt=DEFAULT_DATEFMT)
    # logging.basicConfig()

# TODO: make account for the non-printing characters
CONSOLE_COLS = int(shutil.get_terminal_size(fallback=(2e3, 24)).columns - 1)
# HH:MM:SS.sss vs YYYY-MM-DDTHH:MM:SS.ssssss, shorter format for the console
if DEV_MODE:  # Do this to have longer-format console messages
    CONSOLE_FMT = "%(date)sT%(time)s " + f"%(message).{CONSOLE_COLS - 27}s"
else:
    CONSOLE_FMT = "%(time).12s " + f"%(message).{CONSOLE_COLS - 13}s"
PKT_LOG_FMT = "%(date)sT%(time)s %(_packet)s"

# How to strip ASCII colour from a text file:
#   sed -r "s/\x1B\[(([0-9]{1,2})?(;)?([0-9]{1,2})?)?[m,K,H,f,J]//g" file_name

# used with packet logging
BANDW_SUFFIX = "%(error_text)s%(comment)s"
COLOR_SUFFIX = "%(red)s%(error_text)s%(cyan)s%(comment)s"

LOG_COLOURS = {
    "DEBUG": "white",
    "INFO": "green",
    "WARNING": "yellow",
    "ERROR": "bold_red",
    "CRITICAL": "bold_red",
}  # default_log_colors

_LOGGER = logging.getLogger(__name__)
if DEV_MODE:
    _LOGGER.setLevel(logging.DEBUG)  # DEBUG may have too much detail

_PKT_LOGGER = logging.getLogger(f"{__name__}_log")

if not _use_color_:
    _LOGGER.warning("Consider installing the colorlog library for colored output")


class StdErrFilter(logging.Filter):  # record.levelno >= logging.WARNING
    """For sys.stderr, process only wanted packets."""

    def filter(self, record) -> bool:
        """Return True if the record is to be processed."""
        return record.levelno >= logging.WARNING


class StdOutFilter(logging.Filter):  # record.levelno < logging.WARNING
    """For sys.stdout, process only wanted packets."""

    def filter(self, record) -> bool:
        """Return True if the record is to be processed."""
        return record.levelno < logging.WARNING


class FileFilter(logging.Filter):  # record.levelno in (logging.INFO, logging.WARNING)
    """For packet logs file, process only wanted packets."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.prev_record = None

    def filter(self, record) -> bool:
        """Return True if the record is to be processed."""
        return record.levelno in (logging.INFO, logging.WARNING)

        # HACK: to stop duplicate logging...
        if not hasattr(record, "dtm"):
            self.prev_record = None
        elif self.prev_record:
            if self.prev_record.dtm != record.dtm:
                record, self.prev_record = self.prev_record, record
        else:
            self.prev_record = record
            return False

        return record.levelno in (logging.INFO, logging.WARNING)


def set_pkt_logging(logger=_PKT_LOGGER, cc_stdout=False, **kwargs) -> None:
    """Create/configure handlers, formatters, etc.

    Parameters:
    - backup_count: keep this many copies, and rotate at midnight unless...
    - max_bytes: rotate log files when log > rotate_size
    """

    logger.propagate = False  # this is a distinct log from any app/debug logging
    logger.setLevel(logging.INFO)

    file_name = kwargs.get(LOG_FILE_NAME, 0)
    backup_count = kwargs.get(LOG_ROTATE_COUNT, 0)
    max_bytes = kwargs.get(LOG_ROTATE_BYTES, None)

    if _use_color_:
        cons_fmt = colorlog.ColoredFormatter(
            f"%(log_color)s{CONSOLE_FMT + COLOR_SUFFIX}",
            reset=True,
            log_colors=LOG_COLOURS,
        )
    else:
        cons_fmt = logging.Formatter(fmt=CONSOLE_FMT + BANDW_SUFFIX)

    handler = logging.StreamHandler(stream=sys.stderr)
    handler.setFormatter(cons_fmt)
    handler.setLevel(logging.WARNING)
    handler.addFilter(StdErrFilter())
    logger.addHandler(handler)

    if cc_stdout:
        handler = logging.StreamHandler(stream=sys.stdout)
        handler.setFormatter(cons_fmt)
        handler.setLevel(logging.DEBUG)
        handler.addFilter(StdOutFilter())
        logger.addHandler(handler)

    if not file_name:
        return

    if max_bytes:
        backup_count = backup_count or 2
        handler = logging.handlers.RotatingFileHandler(
            file_name, maxBytes=max_bytes, backupCount=backup_count
        )
    elif backup_count:
        handler = logging.handlers.TimedRotatingFileHandler(
            file_name, when="midnight", backupCount=backup_count
        )
    else:
        handler = logging.FileHandler(file_name)

    handler.setFormatter(logging.Formatter(fmt=PKT_LOG_FMT + BANDW_SUFFIX))
    handler.setLevel(logging.INFO)  # INFO (usually), or DEBUG
    handler.addFilter(FileFilter())
    logger.addHandler(handler)


class Packet:
    """The packet class."""

    def __init__(
        self, pkt_dtm: dt, dtm_str: str, pkt_line: str, raw_pkt_line: str
    ) -> None:
        """Create a packet."""
        self._dtm = pkt_dtm
        self.dtm = dtm_str
        self.date, self.time = self.dtm.split("T")

        self._pkt_str = pkt_line
        self._raw_pkt_str = raw_pkt_line
        self.packet, self.error_text, self.comment = self._split_pkt_line(pkt_line)
        self._packet = self.packet + " " if self.packet else ""  # NOTE: hack 4 logging

        self.addrs = [None] * 3
        self.src_addr = self.dst_addr = None
        self._hdr = None
        self._idx = None

        self._is_valid = None
        if not self.is_valid:
            raise ValueError("not a valid packet")

    def __repr__(self) -> str:
        """Return an unambiguous string representation of this object."""
        return str(self._raw_pkt_str or self._pkt_str)

    def __str__(self) -> str:
        """Return a brief readable string representation of this object."""
        return self.packet or ""

    def __eq__(self, other) -> bool:
        if not hasattr(other, "packet"):
            return NotImplemented
        return self.packet == other.packet

    @staticmethod
    def _split_pkt_line(pkt_line: str) -> Tuple[str, str, str]:
        # dtm_str pkt_str[ < parser-message:][ * evofw3-err_msg][ # evofw3-comment]

        pkt_line, _, pkt_comment = pkt_line.partition("#")
        pkt_line, _, pkt_err_msg = pkt_line.partition("*")
        packet_str, _, _ = pkt_line.partition("<")
        return (
            packet_str.strip(),
            f" * {pkt_err_msg.strip()}" if pkt_err_msg else "",
            f" # {pkt_comment.strip()}" if pkt_comment else "",
        )

    @property
    def is_valid(self) -> Optional[bool]:
        """Return True if a valid packets, otherwise return False/None & log it."""
        # 'good' packets are not logged here, as they may be for silent discarding

        def invalid_addresses() -> bool:
            """Return True if the address fields are invalid (create any addresses)."""
            try:
                self.src_addr, self.dst_addr, self.addrs = extract_addrs(
                    self.packet[11:40]
                )
                # print(extract_addrs.cache_info())
            except CorruptAddrSetError:
                return True

        if self._is_valid is not None or not self._pkt_str:
            return self._is_valid

        self._is_valid = False
        if self.error_text:  # log all packets with an error
            if self.packet:
                _PKT_LOGGER.warning("%s < Bad packet: ", self, extra=self.__dict__)
            else:
                _PKT_LOGGER.warning("< Bad packet: ", extra=self.__dict__)
            return False

        if not self.packet and self.comment:  # log null packets only if has a comment
            _PKT_LOGGER.warning("", extra=self.__dict__)  # best as a debug?
            return False

        # TODO: these packets shouldn't go to the packet log, only STDERR?
        if not MESSAGE_REGEX.match(self.packet):
            err_msg = "invalid packet structure"
        elif int(self.packet[46:49]) * 2 != len(self.packet[50:]):
            err_msg = "mismatched payload length"
        elif invalid_addresses():
            err_msg = "invalid packet addresses"
        else:
            _PKT_LOGGER.info("%s ", self.packet, extra=self.__dict__)
            self._is_valid = True
            return True

        _PKT_LOGGER.debug("%s < Bad packet: %s ", self, err_msg, extra=self.__dict__)
        return False

    @property
    def _header(self) -> Optional[str]:
        """Return the QoS header of this packet."""

        # TODO: a mess: extract_addrs, an expensive function, is possibly called twice
        if self._hdr is None and self.is_valid:
            self._hdr = _pkt_header(self.packet)
        return self._hdr

    @property
    def _index(self) -> Optional[str]:
        """Return the index/ordinal of this packet."""

        if self._idx is None and self.is_valid:
            self._idx = _pkt_header_idx(self.packet)
        return self._idx



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser."""

import logging
from inspect import getmembers, isclass
from sys import modules
from types import SimpleNamespace
from typing import Any, Dict, List, Optional

from .command import FUNC, TIMEOUT, Command, Priority
from .const import (
    _000C_DEVICE,
    ATTR_HEAT_DEMAND,
    ATTR_RELAY_DEMAND,
    ATTR_SETPOINT,
    ATTR_TEMP,
    ATTR_WINDOW_OPEN,
    DEVICE_HAS_BATTERY,
    DEVICE_TABLE,
    DEVICE_TYPES,
    DISCOVER_ALL,
    DISCOVER_PARAMS,
    DISCOVER_SCHEMA,
    DISCOVER_STATUS,
    DOMAIN_TYPE_MAP,
    __dev_mode__,
    id_to_address,
)
from .exceptions import CorruptStateError
from .helpers import dev_id_to_hex, schedule_task
from .opentherm import MSG_ID, MSG_TYPE, VALUE  # R8810A_MSG_IDS
from .ramses import RAMSES_DEVICES

I_, RQ, RP, W_ = " I", "RQ", "RP", " W"

DEFAULT_BDR_ID = "13:000730"
DEFAULT_EXT_ID = "17:000730"
DEFAULT_THM_ID = "03:000730"


DEV_MODE = __dev_mode__ and False

_LOGGER = logging.getLogger(__name__)
if DEV_MODE:
    _LOGGER.setLevel(logging.DEBUG)


DEVICE_CLASS = SimpleNamespace(
    BDR="BDR",  # Electrical relay
    CTL="CTL",  # Controller
    C02="C02",  # HVAC C02 sensor
    DEV="DEV",  # Generic device
    DHW="DHW",  # DHW sensor
    EXT="EXT",  # External weather sensor
    FAN="FAN",  # HVAC fan, 31D[9A]: 20|29|30|37 (some, e.g. 29: only 31D9)
    GWY="GWY",  # Gateway interface (RF to USB), aka HGI
    HUM="HUM",  # HVAC humidity sensor, 1260: 32
    OTB="OTB",  # OpenTherm bridge
    PRG="PRG",  # Programmer
    RFG="RFG",  # RF gateway (RF to ethernet)
    STA="STA",  # Thermostat
    SWI="SWI",  # HVAC switch, 22F[13]: 02|06|20|32|39|42|49|59 (no 20: are both)
    TRV="TRV",  # Thermostatic radiator valve
    UFC="UFC",  # UFH controller
)
_DEV_TYPE_TO_CLASS = {
    None: DEVICE_CLASS.DEV,  # a generic, promotable device
    "00": DEVICE_CLASS.TRV,
    "01": DEVICE_CLASS.CTL,
    "02": DEVICE_CLASS.UFC,
    "03": DEVICE_CLASS.STA,
    "04": DEVICE_CLASS.TRV,
    "07": DEVICE_CLASS.DHW,
    "10": DEVICE_CLASS.OTB,
    "12": DEVICE_CLASS.STA,  # 12: can act like a DEVICE_CLASS.PRG
    "13": DEVICE_CLASS.BDR,
    "17": DEVICE_CLASS.EXT,
    "18": DEVICE_CLASS.GWY,
    "20": DEVICE_CLASS.FAN,
    "22": DEVICE_CLASS.STA,  # 22: can act like a DEVICE_CLASS.PRG
    "23": DEVICE_CLASS.PRG,
    "29": DEVICE_CLASS.FAN,
    "30": DEVICE_CLASS.RFG,  # also: FAN
    "32": DEVICE_CLASS.HUM,  # also: SWI
    "34": DEVICE_CLASS.STA,
    "37": DEVICE_CLASS.FAN,
    "39": DEVICE_CLASS.SWI,
    "42": DEVICE_CLASS.SWI,
    "49": DEVICE_CLASS.SWI,
    "59": DEVICE_CLASS.SWI,
}  # these are the default device classes for common types


class Entity:
    """The Device/Zone base class."""

    def __init__(self, gwy) -> None:
        self._loop = gwy._loop

        self._gwy = gwy
        self.id = None

        self._msgs = {}
        self._msgz = {I_: {}, RQ: {}, RP: {}, W_: {}}

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        pass

    def _get_msg_value(self, code, key=None) -> dict:
        if self._msgs.get(code):
            if isinstance(self._msgs[code].payload, list):
                return self._msgs[code].payload

            if key is not None:
                return self._msgs[code].payload.get(key)

            result = self._msgs[code].payload
            return {
                k: v
                for k, v in result.items()
                if k[:1] != "_" and k not in ("domain_id", "zone_idx")
            }

    def _handle_msg(self, msg) -> None:  # TODO: beware, this is a mess
        if msg.code in self._msgz[msg.verb]:
            self._msgz[msg.verb][msg.code][msg._pkt._index] = msg
        else:
            self._msgz[msg.verb][msg.code] = {msg._pkt._index: msg}

        if msg.verb in (I_, RP):  # TODO: deprecate
            self._msgs[msg.code] = msg

    @property
    def _dump_msgs(self) -> List:
        return [msg for msg in self._msgs.values()]

    def _send_cmd(self, code, dest_id, payload, verb=RQ, **kwargs) -> None:
        self._msgs.pop(code, None)  # remove the old one, so we can tell if RP'd rcvd
        self._gwy.send_cmd(Command(verb, code, payload, dest_id, **kwargs))

    def _msg_payload(self, msg, key=None) -> Optional[Any]:
        if msg and not msg.is_expired:
            if key:
                return msg.payload.get(key)
            return {k: v for k, v in msg.payload.items() if k[:1] != "_"}

    def _msg_expired(self, msg_name: str) -> Optional[bool]:
        attr = f"_{msg_name}"
        if not hasattr(self, attr):
            _LOGGER.error("%s: is not tracking %s msgs", self, msg_name)
            return

        msg = getattr(self, f"_{msg_name}")
        if not msg:
            _LOGGER.warning("%s: has no valid %s msg", self, msg_name)
        # elif msg_name != RAMSES_CODES[msg.code][NAME]:
        #     _LOGGER.warning(
        #         "%s: Message(%s) doesn't match name: %s",
        #         self,
        #         msg._pkt._header,
        #         msg_name,
        #     )
        #     assert False, msg.code
        elif msg.is_expired:
            _LOGGER.warning(
                "%s: Message(%s) has expired (%s)", self, msg._pkt._header, attr
            )
        else:
            return True

    @property
    def _codes(self) -> dict:
        return {
            "codes": sorted([k for k, v in self._msgs.items()]),
        }

    @property
    def controller(self):  # -> Optional[Controller]:
        """Return the entity's controller, if known."""

        return self._ctl  # TODO: if the controller is not known, try to find it?


class DeviceBase(Entity):
    """The Device base class."""

    def __init__(self, gwy, dev_addr, ctl=None, domain_id=None) -> None:
        _LOGGER.debug("Creating a Device: %s (%s)", dev_addr.id, self.__class__)
        super().__init__(gwy)

        self.id = dev_addr.id
        if self.id in gwy.device_by_id:
            raise LookupError(f"Duplicate device: {self.id}")

        gwy.device_by_id[self.id] = self
        gwy.devices.append(self)

        self._ctl = self._set_ctl(ctl) if ctl else None

        self._domain_id = domain_id
        self._parent = None

        self.addr = dev_addr
        self.hex_id = dev_id_to_hex(dev_addr.id)
        self.type = dev_addr.type

        if self.type in DEVICE_TABLE:
            self._has_battery = DEVICE_TABLE[self.addr.type].get("has_battery")
            self._is_actuator = DEVICE_TABLE[self.addr.type].get("is_actuator")
            self._is_sensor = DEVICE_TABLE[self.addr.type].get("is_sensor")
        else:
            self._has_battery = None
            self._is_actuator = None
            self._is_sensor = None

    def __repr__(self) -> str:
        return f"{self.id} ({self._domain_id})"

    def __str__(self) -> str:
        return f"{self.id} ({DEVICE_TYPES.get(self.id[:2])})"

    def __lt__(self, other) -> bool:
        if not hasattr(other, "id"):
            return NotImplemented
        return self.id < other.id

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        # sometimes, battery-powered devices will respond to an RQ (e.g. bind mode)
        # super()._discover(discover_flag=discover_flag)

        if discover_flag & DISCOVER_SCHEMA and self.type not in DEVICE_HAS_BATTERY:
            self._send_cmd("1FC9", retries=3)  # rf_bind

        # if discover_flag & DISCOVER_PARAMS and self.type not in DEVICE_HAS_BATTERY:
        #     pass

        if discover_flag & DISCOVER_STATUS and self.type not in DEVICE_HAS_BATTERY:
            self._send_cmd("0016", retries=3)  # rf_check

    def _send_cmd(self, code, **kwargs) -> None:
        dest = kwargs.pop("dest_addr", self.id)
        payload = kwargs.pop("payload", "00")
        super()._send_cmd(code, dest, payload, **kwargs)

    def _set_ctl(self, ctl) -> None:  # self._ctl
        """Set the device's parent controller, after validating it."""

        if self._ctl is ctl:
            return
        if self._ctl is not None:
            raise CorruptStateError(f"{self} changed controller: {self._ctl} to {ctl}")

        #  I --- 01:078710 --:------ 01:144246 1F09 003 FF04B5  # has been seen
        if not isinstance(ctl, Controller) and not ctl._is_controller:
            raise TypeError(f"Device {ctl} is not a controller")

        self._ctl = ctl
        ctl.device_by_id[self.id] = self
        ctl.devices.append(self)
        _LOGGER.debug("%s: controller now set to %s", self, ctl)

    def _handle_msg(self, msg) -> None:
        """Check that devices only handle messages they have sent."""
        assert msg.src is self, "Devices should only keep msgs they sent"
        super()._handle_msg(msg)

    @property
    def _is_present(self) -> bool:
        """Try to exclude ghost devices (as caused by corrupt packet addresses)."""
        return any(
            m.src == self for m in self._msgs.values() if not m.is_expired
        )  # TODO: needs addressing

    @property
    def _is_controller(self) -> Optional[bool]:  # 1F09
        if self._ctl is not None:
            return self._ctl is self

        # if isinstance(device, Controller):
        # if domain_id == "FF"
        # if dev_addr.type in SYSTEM_CLASSES:
        if self.type in ("01", "23"):
            pass
        # if "1F09" in self._msgs:  # TODO: needs to add msg as attr
        #     return self._msgs["1F09"].verb == I_
        # if "31D9" in self._msgs:  # TODO: needs to add msg as attr
        #     return self._msgs["31D9"].verb == I_
        return False

    @property
    def schema(self) -> dict:
        """Return the fixed attributes of the device (e.g. TODO)."""

        return self._codes if DEV_MODE else {}

    @property
    def params(self):
        return {}

    @property
    def status(self):
        return {}


class Actuator:  # 3EF0, 3EF1

    ACTUATOR_CYCLE = "actuator_cycle"
    ACTUATOR_ENABLED = "actuator_enabled"  # boolean
    ACTUATOR_STATE = "actuator_state"
    ENABLED = "enabled"
    MODULATION_LEVEL = "modulation_level"

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        super()._discover(discover_flag=discover_flag)

        if discover_flag & DISCOVER_STATUS:
            self._send_cmd("3EF1")  # No RPs to 3EF0

    def _handle_msg(self, msg) -> None:
        super()._handle_msg(msg)

        if msg.code == "3EF0" and msg.verb == I_:  # NOT RP, TODO: why????
            self._send_cmd("3EF1", priority=Priority.LOW, retries=1)

    @property
    def actuator_cycle(self) -> Optional[dict]:  # 3EF1
        return self._msg_payload(self._msgs.get("3EF1"))

    @property
    def actuator_state(self) -> Optional[dict]:  # 3EF0 (mod_level, flame_active, etc.)
        return self._msg_payload(self._msgs.get("3EF0"))

    @property
    def enabled(self) -> Optional[bool]:  # 3EF0, 3EF1
        """Return the actuator's current state."""
        msgs = [m for m in self._msgs.values() if m.code in ("3EF0", "3EF1")]
        return max(msgs).payload[self.ACTUATOR_ENABLED] if msgs else None

    @property
    def modulation_level(self) -> Optional[float]:  # 3EF0/3EF1
        msgs = [m for m in self._msgs.values() if m.code in ("3EF0", "3EF1")]
        return max(msgs).payload[self.MODULATION_LEVEL] if msgs else None

    @property
    def status(self) -> dict:
        return {
            **super().status,
            self.ACTUATOR_CYCLE: self.actuator_cycle,
            self.ACTUATOR_STATE: self.actuator_state,
            self.MODULATION_LEVEL: self.modulation_level,  # TODO: keep? (is duplicate)
        }


class BatteryState:  # 1060

    BATTERY_LOW = "battery_low"  # boolean
    BATTERY_STATE = "battery_state"  # percentage

    @property
    def battery_low(self) -> Optional[bool]:  # 1060
        if "1060" in self._msgs:
            return self._msgs["1060"].payload[self.BATTERY_LOW]

    @property
    def battery_state(self) -> Optional[dict]:  # 1060
        return self._msg_payload(self._msgs.get("1060"))

    @property
    def status(self) -> dict:
        return {
            **super().status,
            self.BATTERY_STATE: self.battery_state,
        }


class HeatDemand:  # 3150

    HEAT_DEMAND = ATTR_HEAT_DEMAND  # percentage valve open

    @property
    def heat_demand(self) -> Optional[float]:  # 3150
        if "3150" in self._msgs:
            return self._msgs["3150"].payload[self.HEAT_DEMAND]

    @property
    def status(self) -> dict:
        return {
            **super().status,
            self.HEAT_DEMAND: self.heat_demand,
        }


class Setpoint:  # 2309

    SETPOINT = ATTR_SETPOINT  # degrees Celsius

    @property
    def setpoint(self) -> Optional[float]:  # 2309
        try:
            if "2309" in self._msgs:
                return self._msgs["2309"].payload[self.SETPOINT]
        except TypeError:  # FIXME: 12: as a controller = {[{}, ...]}, not {}
            pass

    @property
    def status(self) -> dict:
        return {
            **super().status,
            self.SETPOINT: self.setpoint,
        }


class Temperature:  # 30C9 (fakeable)

    TEMPERATURE = ATTR_TEMP  # degrees Celsius

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._30C9_faked = None

    def _make_fake(self, bind=None):
        self._30C9_faked = True
        if bind:
            self._bind()
        _LOGGER.error("%s: Faking now enabled", self)  # TODO: shoudl be info/debug

    def _bind(self):
        if not self._30C9_faked:
            raise TypeError("Can't bind sensor (Faking is not enabled)")

    @property
    def temperature(self) -> Optional[float]:  # 30C9
        try:
            if "30C9" in self._msgs:
                return self._msgs["30C9"].payload[self.TEMPERATURE]
        except TypeError:  # FIXME: 12: as a controller = {[{}, ...]}, not {}
            pass

    @temperature.setter
    def temperature(self, value) -> None:  # 30C9
        if not self._30C9_faked:
            raise AttributeError("Can't set attribute (Faking is not enabled)")

        cmd = Command.put_sensor_temp(self.id, value)
        # cmd = Command.put_sensor_temp(
        #     self._gwy.rfg.id if self == self._gwy.rfg._faked_thm else self.id, value
        # )
        self._gwy.send_cmd(cmd)

    @property
    def status(self) -> dict:
        return {
            **super().status,
            self.TEMPERATURE: self.temperature,
        }


class DeviceInfo:  # 10E0

    RF_BIND = "rf_bind"
    DEVICE_INFO = "device_info"

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        if discover_flag & DISCOVER_SCHEMA and self.type not in DEVICE_HAS_BATTERY:
            self._send_cmd("1FC9", retries=3)  # rf_bind
            if self.type != "13":
                self._send_cmd("10E0", retries=3)  # TODO: use device hints

    @property
    def device_info(self) -> Optional[dict]:  # 10E0
        return self._msg_payload(self._msgs.get("10E0"))

    @property
    def schema(self) -> dict:
        result = super().schema
        # result.update({self.RF_BIND: self._msg_payload(self._msgs.get("1FC9"))})
        if "10E0" in self._msgs or "10E0" in RAMSES_DEVICES.get(self.type, []):
            result.update({self.DEVICE_INFO: self.device_info})
        return result


class Device(DeviceInfo, DeviceBase):
    """The Device base class - also used for unknown device types."""

    __dev_class__ = DEVICE_CLASS.DEV  # DEVICE_TYPES = ("??", )

    def _handle_msg(self, msg) -> None:
        super()._handle_msg(msg)

        if not msg._gwy.config.enable_eavesdrop:
            return

        if self._ctl is not None and "parent_idx" in msg.payload:
            # TODO: is buggy - remove? how?
            self._set_parent(self._ctl._evo._get_zone(msg.payload["parent_idx"]))

    def _set_parent(self, parent, domain=None) -> None:  # self._parent
        """Set the device's parent zone, after validating it.

        There are three possible sources for the parent zone of a device:
        1. a 000C packet (from their controller) for actuators only
        2. a message.payload["zone_idx"]
        3. the sensor-matching algorithm for zone sensors only

        Devices don't have parents, rather: Zones have children; a mis-configured
        system could have a device as a child of two domains.
        """

        # these imports are here to prevent circular references
        from .systems import System
        from .zones import DhwZone, Zone

        if self._parent is not None and self._parent is not parent:
            raise CorruptStateError(
                f"{self} changed parent: {self._parent} to {parent}, "
            )

        if isinstance(parent, Zone):
            if domain and domain != parent.idx:
                raise TypeError(f"{self}: domain must be {parent.idx}, not {domain}")
            domain = parent.idx

        elif isinstance(parent, DhwZone):  # usu. FA
            if domain not in ("F9", "FA"):  # may not be known if eavesdrop'd
                raise TypeError(f"{self}: domain must be F9 or FA, not {domain}")

        elif isinstance(parent, System):  # usu. FC
            if domain != "FC":  # was: not in ("F9", "FA", "FC", "HW"):
                raise TypeError(f"{self}: domain must be FC, not {domain}")

        else:
            raise TypeError(f"{self}: parent must be System, DHW or Zone, not {parent}")

        self._set_ctl(parent._ctl)
        self._parent = parent
        self._domain_id = domain

        if hasattr(parent, "devices") and self not in parent.devices:
            parent.devices.append(self)
            parent.device_by_id[self.id] = self
        _LOGGER.debug("Device %s: parent now set to %s", self, parent)

    @property
    def zone(self) -> Optional[Entity]:  # should be: Optional[Zone]
        """Return the device's parent zone, if known."""

        return self._parent

    @property
    def has_battery(self) -> Optional[bool]:  # 1060
        """Return True if a device is battery powered (excludes battery-backup)."""

        if self._has_battery is not None:
            return self._has_battery

        if "1060" in self._msgs:
            self._has_battery = True

        return self._has_battery

    @property
    def schema(self) -> dict:
        """Return the fixed attributes of the device (e.g. TODO)."""

        return {
            **super().schema,
            "dev_class": self.__dev_class__,
        }


class RfiGateway(DeviceBase):  # GWY: 18
    """The HGI80 base class."""

    __dev_class__ = DEVICE_CLASS.GWY  # DEVICE_TYPES = ("18", )

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._ctl = None
        self._domain_id = "FF"
        self._evo = None

        self._faked_bdr = None
        self._faked_ext = None
        self._faked_thm = None

    def _handle_msg(self, msg) -> None:
        def fake_addrs(msg, faked_dev):
            msg.src == faked_dev if msg.src is self else self
            msg.dst == faked_dev if msg.dst is self else self
            return msg

        super()._handle_msg(msg)

        # the following is for aliased devices (not fully-faked devices)
        if msg.code in ("3EF0",) and self._faked_bdr:
            self._faked_bdr._handle_msg(fake_addrs(msg, self._faked_bdr))

        if msg.code in ("0002",) and self._faked_ext:
            self._faked_ext._handle_msg(fake_addrs(msg, self._faked_ext))

        if msg.code in ("30C9",) and self._faked_thm:
            self._faked_thm._handle_msg(fake_addrs(msg, self._faked_thm))

    def _create_fake_dev(self, dev_type, device_id) -> Device:
        if device_id[:2] != dev_type:
            raise TypeError(f"Invalid device ID {device_id} for type '{dev_type}:'")

        dev = self.device_by_id.get(device_id)
        if dev:
            _LOGGER.warning("Destroying %s", dev)
            if dev._ctl:
                del dev._ctl.device_by_id[dev.id]
                dev._ctl.devices.remove(dev)
                dev._ctl = None
            del self.device_by_id[dev.id]
            self.devices.remove(dev)
            dev = None

        dev = self._get_device(id_to_address(device_id))
        dev._make_fake(bind=True)
        return dev

    def create_fake_bdr(self, device_id=DEFAULT_BDR_ID) -> Device:
        """Bind a faked relay (BDR91A) to a controller (i.e. to a domain/zone).

        Will alias the gateway (as "13:000730"), or create a fully-faked 13:.

        HGI80s can only alias one device of a type (use_gateway), but evofw3-based
        gateways can also fully fake multiple devices of the same type.
        """
        if device_id in (self.id, None):
            device_id = DEFAULT_BDR_ID
        device = self._create_fake_dev("13", device_id=device_id)

        if device.id == DEFAULT_BDR_ID:
            self._faked_bdr = device
        return device

    def create_fake_ext(self, device_id=DEFAULT_EXT_ID) -> Device:
        """Bind a faked external sensor (???) to a controller.

        Will alias the gateway (as "17:000730"), or create a fully-faked 17:.

        HGI80s can only alias one device of a type (use_gateway), but evofw3-based
        gateways can also fully fake multiple devices of the same type.
        """

        if device_id in (self.id, None):
            device_id = DEFAULT_EXT_ID
        device = self._create_fake_dev("17", device_id=device_id)

        if device.id == DEFAULT_EXT_ID:
            self._faked_ext = device
        return device

    def create_fake_thm(self, device_id=DEFAULT_THM_ID) -> Device:
        """Bind a faked zone sensor (TR87RF) to a controller (i.e. to a zone).

        Will alias the gateway (as "03:000730"), or create a fully-faked 34:, albeit
        named "03:xxxxxx".

        HGI80s can only alias one device of a type (use_gateway), but evofw3-based
        gateways can also fully fake multiple devices of the same type.
        """
        if device_id in (self.id, None):
            device_id = DEFAULT_THM_ID
        device = self._create_fake_dev("03", device_id=device_id)

        if device.id == DEFAULT_THM_ID:
            self._faked_thm = device
        return device

    @property
    def schema(self):
        return {
            "device_id": self.id,
            "faked_bdr": self._faked_bdr and self._faked_bdr.id,
            "faked_ext": self._faked_ext and self._faked_ext.id,
            "faked_thm": self._faked_thm and self._faked_thm.id,
        }

    @property
    def params(self):
        return {}

    @property
    def status(self):
        return {}


class Controller(Device):  # CTL (01):
    """The Controller base class."""

    __dev_class__ = DEVICE_CLASS.CTL  # DEVICE_TYPES = ("01", )

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._ctl = self  # or args[1]
        self._domain_id = "FF"
        self._evo = None

        self.devices = []  # [self]
        self.device_by_id = {}  # {self.id: self}

    # def _discover(self, discover_flag=DISCOVER_ALL) -> None:
    #     super()._discover(discover_flag=discover_flag)

    #     if discover_flag & DISCOVER_SCHEMA and self.type not in DEVICE_HAS_BATTERY:
    #         pass  # self._send_cmd("1F09", retries=3)

    # #     if discover_flag & DISCOVER_STATUS and self.type not in DEVICE_HAS_BATTERY:
    # #         self._send_cmd("0016", retries=3)  # rf_check


class Programmer(Controller):  # PRG (23):
    """The Controller base class."""

    __dev_class__ = DEVICE_CLASS.PRG  # DEVICE_TYPES = ("23", )


class UfhController(Device):  # UFC (02):
    """The UFC class, the HCE80 that controls the UFH zones."""

    __dev_class__ = DEVICE_CLASS.UFC  # DEVICE_TYPES = ("02", )

    HEAT_DEMAND = ATTR_HEAT_DEMAND

    # 12:27:24.398 067  I --- 02:000921 --:------ 01:191718 3150 002 0360
    # 12:27:24.546 068  I --- 02:000921 --:------ 01:191718 3150 002 065A
    # 12:27:24.693 067  I --- 02:000921 --:------ 01:191718 3150 002 045C
    # 12:27:24.824 059  I --- 01:191718 --:------ 01:191718 3150 002 FC5C
    # 12:27:24.857 067  I --- 02:000921 --:------ 02:000921 3150 006 0060-015A-025C

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.klass = self.__dev_class__

        self._circuits = {}
        self._setpoints = None
        self._heat_demand = None

        self.devices = []  # [self]
        self.device_by_id = {}  # {self.id: self}

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        super()._discover(discover_flag=discover_flag)

        if discover_flag & DISCOVER_SCHEMA:
            [  # 000C: used to find evo zone for each configured channel
                self._send_cmd("000C", payload=f"{idx:02X}{_000C_DEVICE.UFH}")
                for idx in range(8)  # for each possible UFH channel/circuit
            ]

        # if discover_flag & DISCOVER_PARAMS:
        #     pass

        # if discover_flag & DISCOVER_STATUS:
        #     pass

        # [  # 3150: no answer
        #     self._send_cmd("3150", payload=f"{zone_idx:02X}")for zone_idx in range(8)
        # ]

        # [  # 22C9: no answer
        #     self._send_cmd("22C9", payload=f"{payload}")
        #     for payload in ("00", "0000", "01", "0100")
        # ]

        # [  # 22D0: dunno, always: {'unknown': '000002'}
        #     self._send_cmd("22D0", payload=f"{payload}")
        #     for payload in ("00", "0000", "00000002")
        # ]

        [  # 0005: shows which channels are active - ?no use? (see above)
            self._send_cmd("0005", payload=f"00{zone_type}")
            for zone_type in ("09",)  # _0005_ZONE_TYPE, also ("00", "04", "0F")
            # for zone_type in _0005_ZONE_TYPE
        ]

    def _handle_msg(self, msg) -> None:
        super()._handle_msg(msg)

        if msg.code == "000C":
            assert "ufh_idx" in msg.payload, "wsdfh"
            if msg.payload["zone_id"] is not None:
                self._circuits[msg.payload["ufh_idx"]] = msg

        elif msg.code == "22C9":
            if isinstance(msg.payload, list):
                self._setpoints = msg
            # else:
            #     pass  # update the self._circuits[]

        elif msg.code == "3150":
            if isinstance(msg.payload, list):
                self._heat_demands = msg
            elif "domain_id" in msg.payload:
                self._heat_demand = msg
            # else:
            #     pass  # update the self._circuits[]

        # "0008|FA/FC", "22C9|array", "22D0|none", "3150|ZZ/array(/FC?)"

    @property
    def circuits(self) -> Optional[Dict]:  # 000C
        return {
            k: {"zone_idx": m.payload["zone_id"]} for k, m in self._circuits.items()
        }

        # def fix(k):
        #     return "zone_idx" if k == "zone_id" else k

        # return [
        #     {fix(k): v for k, v in m.payload.items() if k in ("ufh_idx", "zone_id")}
        #     for m in self._circuits.values()
        # ]

    @property
    def heat_demand(self) -> Optional[float]:  # 3150
        if self._heat_demand:
            return self._heat_demand.payload[self.HEAT_DEMAND]

    @property
    def relay_demand(self) -> Optional[Dict]:  # 0008
        try:
            return self._msgs["0008"].payload[ATTR_RELAY_DEMAND]
        except KeyError:
            return

    @property
    def setpoints(self) -> Optional[Dict]:  # 22C9
        if self._setpoints is None:
            return

        return {
            c["ufh_idx"]: {"temp_high": c["temp_high"], "temp_low": c["temp_low"]}
            for c in self._setpoints.payload
        }
        # return [
        #     {k: v for k, v in d.items() if k in ("ufh_idx", "temp_high", "temp_low")}
        #     for d in (self._setpoints.payload if self._setpoints else [])
        # ]

    @property  # id, type
    def schema(self) -> dict:
        return {
            **super().schema,
            "circuits": self.circuits,
        }

    @property  # setpoint, config, mode (not schedule)
    def params(self) -> dict:
        return {
            **super().params,
            "circuits": self.setpoints,
        }

    @property
    def status(self) -> dict:
        return {
            **super().status,
            ATTR_HEAT_DEMAND: self.heat_demand,
            ATTR_RELAY_DEMAND: self.relay_demand,
        }


class DhwSensor(BatteryState, Device):  # DHW (07): 10A0, 1260
    """The DHW class, such as a CS92."""

    __dev_class__ = DEVICE_CLASS.DHW  # DEVICE_TYPES = ("07", )

    DHW_PARAMS = "dhw_params"
    TEMPERATURE = ATTR_TEMP
    # _STATE = TEMPERATURE

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._domain_id = "FA"

    def __repr__(self) -> str:
        return f"{self.id} ({self._domain_id}): {self.temperature}"

    @property
    def dhw_params(self) -> Optional[dict]:  # 10A0
        return self._msg_payload(self._msgs.get("10A0"))

    @property
    def temperature(self) -> Optional[float]:  # 1260
        if "1260" in self._msgs:
            return self._msgs["1260"].payload[self.TEMPERATURE]

    @property
    def params(self) -> dict:
        return {
            **super().params,
            self.DHW_PARAMS: self.dhw_params,
        }

    @property
    def status(self) -> dict:
        return {
            **super().status,
            self.TEMPERATURE: self.temperature,
        }


class ExtSensor(Device):  # EXT: 17
    """The EXT class (external sensor), such as a HB85/HB95."""

    __dev_class__ = DEVICE_CLASS.EXT  # DEVICE_TYPES = ("17", )

    LUMINOSITY = "luminosity"  # lux
    TEMPERATURE = "temperature"  # Celsius
    WINDSPEED = "windspeed"  # km/h

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._0002_faked = None
        self._1fc9_state = None

    def __repr__(self) -> str:
        return f"{self.id} ({self._domain_id}): {self.temperature}"

    def _make_fake(self, bind=None):
        self._0002_faked = True
        if bind:
            self._bind()

    def _bind(self):
        def bind_callback(msg) -> None:
            self._1fc9_state == "bound"

            self._gwy._get_device(self, ctl_addr=id_to_address(msg.payload[0][2]))
            self._ctl._evo._get_zone(msg.payload[0][0])._set_sensor(self)

            cmd = Command(
                I_, "1FC9", f"002309{self.hex_id}", self._ctl.id, from_id=self.id
            )
            self._gwy.send_cmd(cmd)

        if not self._0002_faked:
            raise TypeError("Can't bind sensor (Faking is not enabled)")
        self._1fc9_state = "binding"

        cmd = Command.packet(
            I_,
            "1FC9",
            f"000002{self.hex_id}",
            addr0=self.id,
            addr2=self.id,
            callback={FUNC: bind_callback, TIMEOUT: 3},
        )
        self._gwy.send_cmd(cmd)

    @property
    def temperature(self) -> Optional[float]:  # 0002
        if "0002" in self._msgs:
            return self._msgs["0002"].payload[self.TEMPERATURE]

    @temperature.setter
    def temperature(self, value) -> None:  # 0002
        if not self._0002_faked:
            raise AttributeError("Can't set attribute (Faking is not enabled)")

        cmd = Command.put_outdoor_temp(
            self._gwy.rfg.id if self == self._gwy.rfg._faked_ext else self.id, value
        )
        self._gwy.send_cmd(cmd)

    @property
    def luminosity(self) -> Optional[float]:  # 0002
        raise NotImplementedError

    @luminosity.setter
    def luminosity(self, value) -> None:  # 0002
        raise NotImplementedError

    @property
    def windspeed(self) -> Optional[float]:  # 0002
        raise NotImplementedError

    @windspeed.setter
    def windspeed(self, value) -> None:  # 0002
        raise NotImplementedError

    @property
    def status(self) -> dict:
        return {
            **super().status,
            self.LUMINOSITY: self.luminosity,
            self.TEMPERATURE: self.temperature,
            self.WINDSPEED: self.windspeed,
        }


class OtbGateway(Actuator, HeatDemand, Device):  # OTB (10): 22D9, 3220
    """The OTB class, specifically an OpenTherm Bridge (R8810A Bridge)."""

    __dev_class__ = DEVICE_CLASS.OTB  # DEVICE_TYPES = ("10", )

    BOILER_SETPOINT = "boiler_setpoint"
    OPENTHERM_STATUS = "opentherm_status"
    # _STATE = super().MODULATION_LEVEL

    SCHEMA_MDG_IDS = (0x7C, 0x7D, 0x7E, 0x7F)
    PARAMS_MDG_IDS = (0x02, 0x03, 0x0E)
    STATUS_MDG_IDS = (
        0x00,
        0x01,
        0x11,
        0x19,
        0x05,
        0x12,
        0x13,
        0x1A,
        0x1B,
        0x1C,
        0x73,
        0x7C,
    )  # TODO: remove 0x7C

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._domain_id = "FC"

        self._opentherm_msg = self._msgz[RP]["3220"] = {}
        self._supported_msg = {}

    def __repr__(self) -> str:
        return f"{self.id} ({self._domain_id}): {self.modulation_level}"  # 3EF0

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        # see: https://www.opentherm.eu/request-details/?post_ids=2944
        super()._discover(discover_flag=discover_flag)

        if discover_flag & DISCOVER_SCHEMA:
            # 7C - Master Opentherm version (is supported?)
            # 7D - Slave Opentherm version (is supported?)
            # 7E - Master Product Type/Version
            # 7F - Slave Product Type/Version
            [
                self._gwy.send_cmd(Command.get_opentherm_data(self.id, m))
                for m in self.SCHEMA_MDG_IDS  # From OT v2.2: version numbers
                if self._supported_msg.get(m) is not False
                and (
                    not self._opentherm_msg.get(m) or self._opentherm_msg[m].is_expired
                )
            ]

        if discover_flag & DISCOVER_PARAMS:
            # 02 - Master configuration  # TODO: does OTB respond to this?
            # 03 - Slave configuration
            # 0E - Max. relative modulation level (%)
            [
                self._gwy.send_cmd(Command.get_opentherm_data(self.id, m))
                for m in self.PARAMS_MDG_IDS
                if self._supported_msg.get(m) is not False
                and (
                    not self._opentherm_msg.get(m) or self._opentherm_msg[m].is_expired
                )
            ]

        if discover_flag & DISCOVER_STATUS:
            self._gwy.send_cmd(Command(RQ, "22D9", "00", self.id))

            # mandatory: msg_ids = {0x00, 0x01, 0x11, 0x19}
            # 00 - Master/Slave status flags
            # 01 - CH water temperature setpoint
            # 11 - Relative modulation level (%)
            # 19 - Boiler water temperature

            # evohome: msg_ids |= {0x05, 0x12, 0x13, 0x1A}
            # 05 - Fault flags & OEM fault code
            # 12 - Central heating water pressure (bar)
            # 13 - DHW flow rate (L/min)
            # 1A - DHW temperature

            # other: msg_ids |= {0x1B, 0x1C, 0x73}
            # 1B - Outside temperature
            # 1C - Return water temperature
            # 73 - OEM diagnostic code (is supported?)

            [
                self._gwy.send_cmd(Command.get_opentherm_data(self.id, m, retries=0))
                for m in self.STATUS_MDG_IDS
                if self._supported_msg.get(m) is not False
                and (
                    not self._opentherm_msg.get(m) or self._opentherm_msg[m].is_expired
                )
            ]

    def _handle_msg(self, msg) -> None:
        super()._handle_msg(msg)

        if msg.code == "1FD4":  # every 30s
            if msg.payload["ticker"] % 60 in (1, 3):
                self._discover(discover_flag=DISCOVER_PARAMS)
            elif msg.payload["ticker"] % 6 in (0, 2):
                self._discover(discover_flag=DISCOVER_STATUS)

        elif msg.code == "3220":  # all are RP
            if msg.payload[MSG_TYPE] == "Unknown-DataId":
                self._supported_msg[msg.payload[MSG_ID]] = False
            # else:
            #     self._supported_msg[msg.payload[MSG_ID]] = True

    def _opentherm_msg_value(self, msg_id) -> Optional[float]:
        try:
            return self._opentherm_msg[f"{msg_id:02X}"].payload[VALUE]
        except KeyError:
            return

    @property
    def boiler_water_temp(self) -> Optional[float]:  # 3220/0x19
        return self._opentherm_msg_value(0x19)

    @property
    def ch_water_pressure(self) -> Optional[float]:  # 3220/0x12
        return self._opentherm_msg_value(0x12)

    @property
    def dhw_flow_rate(self) -> Optional[float]:  # 3220/0x13
        return self._opentherm_msg_value(0x13)

    @property
    def dhw_temp(self) -> Optional[float]:  # 3220/0x1A
        return self._opentherm_msg_value(0x1A)

    @property
    def rel_modulation_level(self) -> Optional[float]:  # 3220/0x11
        return self._opentherm_msg_value(0x11)

    @property
    def return_cv_temp(self) -> Optional[float]:  # 3220/0x1C
        return self._opentherm_msg_value(0x1C)

    @property
    def boiler_setpoint(self) -> Optional[float]:  # 22D9
        try:
            return self._msgs["22D9"].payload[self.BOILER_SETPOINT]
        except KeyError:
            return

    @property
    def opentherm_status(self) -> dict:
        return {
            "boiler_temp": self.boiler_water_temp,
            "ch_pressure": self.ch_water_pressure,
            "dhw_flow_rate": self.dhw_flow_rate,
            "dhw_temp": self.dhw_temp,
            "rel_modulation_level": self.rel_modulation_level,
            "return_cv_temp": self.return_cv_temp,
        }
        # return {
        #     slugify(self._opentherm_msg[msg_id].payload[MSG_NAME]): (
        #         self._opentherm_msg[msg_id].payload[VALUE]
        #     )
        #     for msg_id in (0x11, 0x12, 0x13, 0x19, 0x1A, 0x1C)
        #     if msg_id in self._opentherm_msg
        # }

    @property
    def status(self) -> dict:
        return {
            **super().status,
            self.BOILER_SETPOINT: self.boiler_setpoint,
            self.OPENTHERM_STATUS: self.opentherm_status,
        }


class Thermostat(BatteryState, Setpoint, Temperature, Device):  # THM (..):
    """The THM/STA class, such as a TR87RF."""

    __dev_class__ = DEVICE_CLASS.STA  # DEVICE_TYPES = ("03", "12", "22", "34")

    # _STATE = super().TEMPERATURE

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._1fc9_state = None

    def __repr__(self) -> str:
        return f"{self.id} ({self._domain_id}): {self.temperature}"

    def _bind(self):
        def bind_callback(msg) -> None:
            self._1fc9_state == "bound"

            self._gwy._get_device(self, ctl_addr=id_to_address(msg.payload[0][2]))
            self._ctl._evo._get_zone(msg.payload[0][0])._set_sensor(self)

            cmd = Command(
                I_, "1FC9", f"002309{self.hex_id}", self._ctl.id, from_id=self.id
            )
            self._gwy.send_cmd(cmd)

        super()._bind()
        self._1fc9_state = "binding"

        callback = {FUNC: bind_callback, TIMEOUT: 3}
        payload = "".join(
            f"00{c}{self.hex_id}" for c in ("2309", "30C9", "0008", "1FC9")
        )
        cmd = Command.packet(
            I_, "1FC9", payload, addr0=self.id, addr2=self.id, callback=callback
        )
        self._gwy.send_cmd(cmd)


class BdrSwitch(Actuator, Device):  # BDR (13):
    """The BDR class, such as a BDR91.

    BDR91s can be used in six disctinct modes, including:
    - x2 boiler controller (FC/TPI): either traditional, or newer heat pump-aware
    - x1 electric heat zones (0x/ELE)
    - x1 zone valve zones (0x/VAL)
    - x2 DHW thingys (F9/DHW, FA/DHW)
    """

    __dev_class__ = DEVICE_CLASS.BDR  # DEVICE_TYPES = ("13", )

    RELAY_DEMAND = "relay_demand"  # percentage
    TPI_PARAMS = "tpi_params"
    # _STATE = super().ENABLED, or relay_demand

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._is_tpi = None

        # if kwargs.get("domain_id") == "FC":  # TODO: F9/FA/FC, zone_idx
        #     self._ctl._set_htg_control(self)

    def __repr__(self) -> str:
        return f"{self.id} ({self._domain_id}): {self.enabled}"  # or: relay_demand?

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        """The BDRs have one of six roles:
         - heater relay *or* a heat pump relay (alternative to an OTB)
         - DHW hot water valve *or* DHW heating valve
         - Zones: Electric relay *or* Zone valve relay

        They all seem to respond thus (TODO: heat pump/zone valve relay):
         - all BDR91As will (erractically) RP to these RQs
             0016, 1FC9 & 0008, 1100, 3EF1
         - all BDR91As will *not* RP to these RQs
             0009, 10E0, 3B00, 3EF0
         - a BDR91A will *periodically* send an I/3B00/00C8 if it is the heater relay
        """

        super()._discover(discover_flag=discover_flag)

        # if discover_flag & DISCOVER_SCHEMA:
        #     self._send_cmd("1FC9")  # will include a 3B00 if is a heater_relay

        if discover_flag & DISCOVER_PARAMS:
            self._send_cmd("1100")

        if discover_flag & DISCOVER_STATUS:
            self._send_cmd("0008")

    def _handle_msg(self, msg) -> None:
        super()._handle_msg(msg)

        if msg.code == "3EF0" and msg.verb == I_:  # NOT RP, TODO: why????
            self._send_cmd("0008", priority=Priority.LOW, retries=1)

        # elif msg.code == "1FC9" and msg.verb == RP:
        #     pass  # only a heater_relay will have 3B00

        # elif msg.code == "3B00" and msg.verb == I_:
        #     pass  # only a heater_relay will I/3B00
        #     # for code in ("0008", "3EF1"):
        #     #     self._send_cmd(code, delay=1)

    @property
    def role(self) -> Optional[str]:
        """Return the role of the BDR91A (there are six possibilities)."""

        if self._domain_id in DOMAIN_TYPE_MAP:
            return DOMAIN_TYPE_MAP[self._domain_id]
        elif self._parent:
            return self._parent.heating_type  # TODO: only applies to zones

    @property
    def _role(self) -> Optional[str]:  # TODO: XXX
        """Return the role of the BDR91A (there are six possibilities)."""

        if self._is_tpi is not None:
            return self._is_tpi

        elif "1FC9" in self._msgs and self._msgs["1FC9"].verb == RP:
            if "3B00" in self._msgs["1FC9"].raw_payload:
                self._is_tpi = True

        elif "3B00" in self._msgs and self._msgs["3B00"].verb == I_:
            self._is_tpi = True

        if self._is_tpi:
            self._domain_id = "FC"  # TODO: check is None first
            self._ctl._set_htg_control(self)

        return self._is_tpi

    @property
    def relay_demand(self) -> Optional[float]:  # 0008
        if "0008" in self._msgs:
            return self._msgs["0008"].payload[self.RELAY_DEMAND]

    @property
    def tpi_params_wip(self) -> Optional[dict]:  # 1100
        return self._msg_payload(self._msgs.get("1100"))

    @property
    def params(self) -> dict:
        return {
            **super().params,
            self.TPI_PARAMS: self.tpi_params_wip,
        }

    @property
    def status(self) -> dict:
        return {
            **super().status,
            self.RELAY_DEMAND: self.relay_demand,
        }


class TrvActuator(BatteryState, HeatDemand, Setpoint, Temperature, Device):  # TRV (04):
    """The TRV class, such as a HR92."""

    __dev_class__ = DEVICE_CLASS.TRV  # DEVICE_TYPES = ("00", "04")

    WINDOW_OPEN = ATTR_WINDOW_OPEN  # boolean
    # _STATE = HEAT_DEMAND

    def __repr__(self) -> str:
        return f"{self.id} ({self._domain_id}): {self.heat_demand}"

    @property
    def window_open(self) -> Optional[bool]:  # 12B0
        if "12B0" in self._msgs:
            return self._msgs["12B0"].payload[self.WINDOW_OPEN]

    @property
    def status(self) -> dict:
        return {
            **super().status,
            self.WINDOW_OPEN: self.window_open,
        }


class FanSwitch(BatteryState, Device):  # SWI (39):
    """The FAN (switch) class, such as a 4-way switch.

    The cardinal codes are 22F1, 22F3.
    """

    __dev_class__ = DEVICE_CLASS.SWI  # DEVICE_TYPES = ("39",)

    BOOST_TIMER = "boost_timer"  # minutes, e.g. 10, 20, 30 minutes
    HEATER_MODE = "heater_mode"  # e.g. auto, off
    HEATER_MODES = {9: "off", 10: "auto"}  # TODO:

    FAN_MODE = "fan_mode"  # e.g. low. high
    FAN_MODES = {
        0: "standby",
        1: "auto",
        2: "low",
        3: "medium",
        4: "high",  # a.k.a. boost if timer on
    }
    FAN_RATE = "fan_rate"  # percentage, 0.0 - 1.0

    @property
    def fan_mode(self) -> Optional[str]:
        if "22F1" in self._msgs:
            return self._msgs["22F1"].payload[self.FAN_MODE]

    @property
    def boost_timer(self) -> Optional[int]:
        if "22F3" in self._msgs:
            return self._msgs["22F3"].payload[self.BOOST_TIMER]

    @property
    def status(self) -> dict:
        return {
            **super().status,
            self.FAN_MODE: self.fan_mode,
            self.BOOST_TIMER: self.boost_timer,
        }


class FanDevice(Device):  # FAN (20/37):
    """The Ventilation class.

    The cardinal code are 31D9, 31DA.
    """

    __dev_class__ = DEVICE_CLASS.FAN  # DEVICE_TYPES = ("20", "37")

    @property
    def fan_rate(self) -> Optional[float]:
        msgs = [m for m in self._msgs.values() if m.code in ("31D9", "31DA")]
        return max(msgs).payload["exhaust_fan_speed"] if msgs else None

    @property
    def boost_timer(self) -> Optional[int]:
        if "31DA" in self._msgs:
            return self._msgs["31DA"].payload["remaining_time"]

    @property
    def relative_humidity(self) -> Optional[float]:
        if "31DA" in self._msgs:
            return self._msgs["31DA"].payload["indoor_humidity"]

    @property
    def status(self) -> dict:
        return {
            **super().status,
            "exhaust_fan_speed": self.fan_rate,
            **(
                {
                    k: v
                    for k, v in self._msgs["31D9"].payload.items()
                    if k != "exhaust_fan_speed"
                }
                if "31D9" in self._msgs
                else {}
            ),
            **(
                {
                    k: v
                    for k, v in self._msgs["31DA"].payload.items()
                    if k != "exhaust_fan_speed"
                }
                if "31DA" in self._msgs
                else {}
            ),
        }


class FanSensorHumidity(BatteryState, Device):  # HUM (32) Humidity sensor:
    """The Sensor class for a humidity sensor.

    The cardinal code is 12A0.
    """

    __dev_class__ = DEVICE_CLASS.HUM  # DEVICE_TYPES = ("32")

    REL_HUMIDITY = "relative_humidity"  # percentage
    TEMPERATURE = "temperature"  # celsius
    DEWPOINT_TEMP = "dewpoint_temp"  # celsius

    @property
    def relative_humidity(self) -> Optional[float]:
        if "12A0" in self._msgs:
            return self._msgs["12A0"].payload[self.REL_HUMIDITY]

    @property
    def temperature(self) -> Optional[float]:
        if "12A0" in self._msgs:
            return self._msgs["12A0"].payload[self.TEMPERATURE]

    @property
    def dewpoint_temp(self) -> Optional[float]:
        if "12A0" in self._msgs:
            return self._msgs["12A0"].payload[self.DEWPOINT_TEMP]

    @property
    def status(self) -> dict:
        return {
            **super().status,
            self.REL_HUMIDITY: self.relative_humidity,
            self.TEMPERATURE: self.temperature,
            self.DEWPOINT_TEMP: self.dewpoint_temp,
        }


CLASS_ATTR = "__dev_class__"
DEVICE_BY_CLASS_ID = {
    getattr(c[1], CLASS_ATTR): c[1]
    for c in getmembers(
        modules[__name__],
        lambda m: isclass(m) and m.__module__ == __name__ and hasattr(m, CLASS_ATTR),
    )
}  # e.g. "CTL": Controller

DEVICE_BY_ID_TYPE = {
    k1: v2
    for k1, v1 in _DEV_TYPE_TO_CLASS.items()
    for k2, v2 in DEVICE_BY_CLASS_ID.items()
    if v1 == k2
}  # e.g. "01": Controller,


def create_device(gwy, dev_addr, dev_class=None, **kwargs) -> Device:
    """Create a device, and optionally perform discovery & start polling."""

    if dev_class is None:
        dev_class = _DEV_TYPE_TO_CLASS.get(dev_addr.type, DEVICE_CLASS.DEV)

    device = DEVICE_BY_CLASS_ID.get(dev_class, Device)(gwy, dev_addr, **kwargs)

    if not gwy.config.disable_discovery:
        schedule_task(device._discover, discover_flag=DISCOVER_SCHEMA)
        schedule_task(device._discover, discover_flag=DISCOVER_PARAMS, delay=14)
        schedule_task(device._discover, discover_flag=DISCOVER_STATUS, delay=15)

    return device
